r"""Script to copy models from the root of the repo"""
import os
import shutil

solutions_dir = os.path.abspath(os.path.dirname(__file__))
files = {
    'tyk1': [
        'models/weather.py',
        'input/intensity.txt',
    ],
    'tyk2': [
        'yamls/co2.yml',
        'models/humidity.py',
        'input/height.txt',
        'input/co2.txt',
        'input/temp.txt',
    ],
    'tyk3': [
        'models/light_v0.py',
        'yamls/light_v0_python.yml',
        'models/weather.py',
        'models/shoot_v1.py',
        'yamls/shoot_v1.yml',
        'yamls/co2.yml',
        'input/height.txt',
        (os.path.join(solutions_dir, 'tyk2/models/co2.py'),
         os.path.join(solutions_dir, 'tyk3/models/co2.py')),
        'meshes/plants-2.obj'
        ],
    'tyk4': [
        'models/light_v0.py',
        'models/weather.py',
        'input/height.txt',
        'meshes/plants-2.obj',
    ],
    'tyk5': [
        'models/shoot_v1.py',
        'meshes/plants-2.obj',
    ],
    'tyk6': [
        'yamls/roots_v1.yml',
        'models/roots_v1.py',
        'yamls/shoot_v3.yml',
        'models/shoot_v3.py',
        'yamls/light_v1_python.yml',
        'models/light_v0.py',
        'yamls/timesync.yml',
        'meshes/plants-2.obj',
    ],
    'tyk7': [
        'models/weather.py',
        'yamls/co2.yml',
        (os.path.join(solutions_dir, 'tyk2/models/co2.py'),
         os.path.join(solutions_dir, 'tyk7/models/co2.py')),
    ],
    'tyk8': [
        'models/light_v0.py',
        'yamls/light_v0_python.yml',
        'input/light_v0.txt',
    ],
}


for k, kset in files.items():
    input_dir = os.path.join(solutions_dir, k, 'input')
    if not os.path.isdir(input_dir):
        os.mkdir(input_dir)
    for f in kset:
        if isinstance(f, tuple):
            src = f[0]
            dst = f[1]
        else:
            src = os.path.join(solutions_dir, '..', f)
            dst = os.path.join(solutions_dir, k, f)
        # print(src, dst)
        assert(os.path.isfile(src))
        dst_dir = os.path.dirname(dst)
        if not os.path.isdir(dst_dir):
            os.mkdir(dst_dir)
        shutil.copy2(src, dst)



import pytest
from yggdrasil import constants
from tests.drivers.test_ConnectionDriver import (
    TestConnectionDriver as base_class)


_comm_types = sorted(
    [x for x in constants.COMPONENT_REGISTRY['comm']['subtypes'].keys()
     if x not in ['value', 'buffer', 'mpi']])


class TestRPCRequestDriver(base_class):
    r"""Test class for RPCRequestDriver class."""

    @pytest.fixture(scope="class")
    def component_subtype(self):
        r"""Subtype of component being tested."""
        return 'rpc_request'

    @pytest.fixture(scope="class", params=_comm_types)
    def ocomm_name(self, request):
        r"""str: Name of the output communicator being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def icomm_name(self, ocomm_name):
        r"""str: Name of the input communicator being tested."""
        return ocomm_name
    
    # @pytest.fixture(scope="class")
    # def timeout(self):
    #     r"""int: Time that should be waiting in seconds."""
    #     return 5.0

    @pytest.fixture(scope="class")
    def route_timeout(self, timeout):
        r"""int: Time to wait for messages to be routed."""
        return 2 * timeout
    
    @pytest.fixture
    def send_comm_kwargs(self, instance, icomm_name):
        r"""dict: Keyword arguments for send comm."""
        out = instance.icomm.opp_comm_kwargs()
        out['request_commtype'] = out['commtype']
        out['commtype'] = 'client'
        return out

    @pytest.fixture
    def recv_comm_kwargs(self, instance):
        r"""dict: Keyword arguments for recv comm."""
        out = instance.ocomm.opp_comm_kwargs()
        out['request_commtype'] = out['commtype']
        out['commtype'] = 'server'
        return out
    
    def test_error_attributes(self, instance):
        r"""Test error raised when trying to access attributes set on recv."""
        err_attr = ['request_id', 'response_address']
        for k in err_attr:
            with pytest.raises(AttributeError):
                getattr(instance, k)

    @pytest.fixture
    def do_send_recv(self, started_instance, send_comm, recv_comm,
                     route_timeout, wait_on_function, nested_approx):
        r"""Perform a send/recv cycle."""
        def do_send_recv_w(msg_send):
            try:
                wait_on_function(lambda: started_instance.is_valid)
                # Send a message to local output
                flag = send_comm.send(msg_send)
                assert(flag)
                # Receive on server side, then send back
                flag, srv_msg = recv_comm.recv(timeout=route_timeout)
                assert(flag)
                assert(srv_msg == nested_approx(msg_send))
                started_instance.printStatus()
                started_instance.printStatus(return_str=True)
                flag = recv_comm.send(srv_msg)
                assert(flag)
                # Receive response on client side
                flag, cli_msg = send_comm.recv(timeout=route_timeout)
                assert(flag)
                assert(cli_msg == nested_approx(msg_send))
            except BaseException:  # pragma: debug
                send_comm.printStatus()
                started_instance.printStatus(verbose=True)
                recv_comm.printStatus()
                raise
        return do_send_recv_w

    def test_send_recv(self, do_send_recv, test_msg):
        r"""Test routing of a short message between client and server."""
        do_send_recv(test_msg)

    def test_send_recv_nolimit(self, do_send_recv, msg_long):
        r"""Test routing of a large message between client and server."""
        do_send_recv(msg_long)



import pytest
import uuid
from yggdrasil.communication import new_comm
from tests.communication.test_CommBase import TestComm as base_class


class TestServerComm(base_class):
    r"""Tests for ServerComm communication class."""

    test_error_send = None
    test_error_recv = None
    test_invalid_direction = None
    test_work_comm = None
    test_send_recv_raw = None

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self):
        r"""Communicator type being tested."""
        return "server"

    @pytest.fixture(scope="class", autouse=True)
    def python_class(self):
        r"""Python class that is being tested."""
        from yggdrasil.communication.ServerComm import ServerComm
        return ServerComm

    def get_send_comm_kwargs(self, commtype, send_comm,
                             testing_options, **kwargs):
        r"""Get keyword arguments for creating a send comm."""
        kwargs.update(direct_connection=True)
        out = super(TestServerComm, self).get_send_comm_kwargs(
            'client', send_comm, testing_options, **kwargs)
        return out

    def test_newcomm_server(self, commtype, close_comm):
        r"""Test creation of server using newcomm."""
        inst = new_comm('testserver_%s' % str(uuid.uuid4()),
                        commtype=commtype)
        close_comm(inst)
        
    def test_call(self, send_comm, recv_comm, testing_options, timeout,
                  wait_on_function):
        r"""Test RPC call."""
        send_comm.maxMsgSize
        recv_comm.maxMsgSize
        send_comm.opp_comms
        recv_comm.opp_comms
        send_comm.sched_task(0.0, send_comm.rpcCall,
                             args=[testing_options['msg']],
                             store_output=True)
        flag, msg_recv = recv_comm.rpcRecv(timeout=timeout)
        assert(flag)
        assert(msg_recv == testing_options['msg'])
        flag = recv_comm.rpcSend(msg_recv)
        assert(flag)
        wait_on_function(lambda: (send_comm.sched_out is not None))
        flag, msg_recv = send_comm.sched_out
        assert(flag)
        assert(msg_recv == testing_options['msg'])

    def test_call_alias(self, send_comm, recv_comm, testing_options,
                        timeout, polling_interval, wait_on_function):
        r"""Test RPC call aliases."""
        # send_comm.sched_task(0.0, send_comm.rpcSend,
        #                      args=[testing_options['msg']],
        #                      store_output=True)
        recv_comm.sched_task(polling_interval, recv_comm.rpcRecv,
                             kwargs=dict(timeout=timeout),
                             store_output=True)
        flag = send_comm.rpcSend(testing_options['msg'])
        assert(flag)
        wait_on_function(lambda: (recv_comm.sched_out is not None))
        flag, msg_recv = recv_comm.sched_out
        # flag, msg_recv = recv_comm.rpcRecv(timeout=timeout)
        assert(flag)
        assert(msg_recv == testing_options['msg'])
        flag = recv_comm.rpcSend(msg_recv)
        assert(flag)
        flag, msg_recv = send_comm.rpcRecv(timeout=timeout)
        assert(flag)
        assert(msg_recv == testing_options['msg'])

    def test_call_nolimit(self, send_comm, recv_comm, msg_long, timeout,
                          wait_on_function):
        r"""Test RPC nolimit call."""
        send_comm.sched_task(0.0, send_comm.call_nolimit,
                             args=[msg_long], store_output=True)
        msg = recv_comm.recv_nolimit(timeout=timeout,
                                     return_message_object=True)
        flag = bool(msg.flag)
        msg_recv = msg.args
        header = msg.header
        assert(flag)
        assert(msg_recv == msg_long)
        assert(isinstance(header, dict))
        flag = recv_comm.send_nolimit(msg_recv)
        assert(flag)
        T = recv_comm.start_timeout()
        while (not T.is_out) and (send_comm.sched_out is None):  # pragma: debug
            recv_comm.sleep()
        recv_comm.stop_timeout()
        flag, msg_recv = send_comm.sched_out
        assert(flag)
        assert(msg_recv == msg_long)

    def test_close_in_thread(self, send_comm, recv_comm):
        r"""Test close of comm in thread."""
        send_comm.close_in_thread()
        recv_comm.close_in_thread()

    def add_filter(self, comm, filter=None, **kwargs):
        r"""Add a filter to a comm.

        Args:
            comm (CommBase): Communication instance to add a filter to.
            filter (FilterBase, optional): Filter class. Defaults to None and is ignored.

        """
        target = comm
        if comm._commtype == 'server':
            target = comm.icomm
        elif comm._commtype == 'client':
            target = comm.ocomm
        return super(TestServerComm, self).add_filter(target, filter=filter,
                                                      **kwargs)
        
    # # This dosn't work for comms that are uni-directional
    # def test_purge_recv(self, send_comm, recv_comm, testing_options,
    #                     wait_on_function):
    #     r"""Test purging messages from the client comm."""
    #     # Purge send while open
    #     flag = send_comm.send(testing_options['msg'])
    #     assert(flag)
    #     wait_on_function(lambda: (recv_comm.n_msg != 0))
    #     assert(recv_comm.n_msg == 1)
    #     send_comm.purge()
    #     assert(send_comm.n_msg == 0)
    #     assert(recv_comm.n_msg == 0)
    #     # Purge send while closed
    #     send_comm.close()
    #     send_comm.purge()



"""This module provides tools for running models using yggdrasil."""
import sys
import os
import time
import copy
import signal
import atexit
from pprint import pformat
from itertools import chain
import socket
from collections import OrderedDict
from yggdrasil.tools import YggClass
from yggdrasil.config import ygg_cfg, cfg_environment, temp_config
from yggdrasil import platform, yamlfile
from yggdrasil.drivers import create_driver
from yggdrasil.components import import_component
from yggdrasil.multitasking import MPI
from yggdrasil.drivers.DuplicatedModelDriver import DuplicatedModelDriver
from yggdrasil.drivers.ModelDriver import ModelDriver


COLOR_TRACE = '\033[30;43;22m'
COLOR_NORMAL = '\033[0m'


class IntegrationError(BaseException):
    r"""Error raised when there is an error in an integration."""
    pass


class YggFunction(YggClass):
    r"""This class wraps function-like behavior around a model.

    Args:
        model_yaml (str, list): Full path to one or more YAML specification
            files containing information defining a partial integration. If
            service_address is set, this should be the name of a service
            registered with the service manager running at the provided
            address.
        service_address (str, optional): Address for service manager that is
            capable of running the specified integration. Defaults to None
            and is ignored.
        **kwargs: Additional keyword arguments are passed to the YggRunner
            constructor.

    Attributes:
        outputs (dict): Input channels providing access to model output.
        inputs (dict): Output channels providing access to model input.
        runner (YggRunner): Runner for model.

    """
    
    def __init__(self, model_yaml, service_address=None, **kwargs):
        import uuid
        from yggdrasil.languages.Python.YggInterface import (
            YggInput, YggOutput, YggRpcClient)
        super(YggFunction, self).__init__()
        # Create and start runner in another process
        self.dummy_name = 'func' + str(uuid.uuid4()).split('-')[0]
        kwargs['complete_partial'] = self.dummy_name
        if service_address:
            # Temporary YAML describing the service
            contents = (f'service:\n'
                        f'    name: {model_yaml}\n'
                        f'    address: {service_address}\n')
            model_yaml = os.path.join(os.getcwd(), self.dummy_name + '.yml')
            with open(model_yaml, 'w') as fd:
                fd.write(contents)
        self.runner = YggRunner(model_yaml, **kwargs)
        # Start the drivers
        self.runner.run()
        self.model_driver = self.runner.modeldrivers[self.dummy_name]
        for k in self.runner.modeldrivers.keys():
            if k != self.dummy_name:
                self.__name__ = k
                break
        self.debug("run started")
        # Create input/output channels
        self.inputs = {}
        self.outputs = {}
        # import zmq; ctx = zmq.Context()
        self.old_environ = os.environ.copy()
        for drv in self.model_driver['input_drivers']:
            for env in drv['instance'].model_env.values():
                os.environ.update(env)
            channel_name = drv['instance'].ocomm.name
            var_name = drv['name'].split('function_')[-1]
            self.outputs[var_name] = drv.copy()
            self.outputs[var_name]['comm'] = YggInput(
                channel_name, no_suffix=True)  # context=ctx)
            if 'vars' in drv['inputs'][0]:
                self.outputs[var_name]['vars'] = drv['inputs'][0]['vars']
            else:
                self.outputs[var_name]['vars'] = [var_name]
        for drv in self.model_driver['output_drivers']:
            for env in drv['instance'].model_env.values():
                os.environ.update(env)
            channel_name = drv['instance'].icomm.name
            var_name = drv['name'].split('function_')[-1]
            self.inputs[var_name] = drv.copy()
            if drv['instance']._connection_type == 'rpc_request':
                self.inputs[var_name]['comm'] = YggRpcClient(
                    channel_name, no_suffix=True)
                self.outputs[var_name] = drv.copy()
                self.outputs[var_name]['comm'] = self.inputs[var_name]['comm']
                if drv['outputs'][0].get('server_replaces', False):
                    srv = drv['outputs'][0]['server_replaces']
                    self.inputs[var_name]['vars'] = srv['input']['vars']
                    self.outputs[var_name]['vars'] = srv['output']['vars']
            else:
                self.inputs[var_name]['comm'] = YggOutput(
                    channel_name, no_suffix=True)  # context=ctx)
                if 'vars' in drv['outputs'][0]:
                    self.inputs[var_name]['vars'] = drv['outputs'][0]['vars']
                else:
                    self.inputs[var_name]['vars'] = [var_name]
        self.debug('inputs: %s, outputs: %s',
                   list(self.inputs.keys()),
                   list(self.outputs.keys()))
        self._stop_called = False
        atexit.register(self.stop)
        # Ensure that vars are strings
        for k, v in chain(self.inputs.items(), self.outputs.items()):
            v_vars = []
            for iv in v['vars']:
                if isinstance(iv, dict):
                    if not iv.get('is_length_var', False):
                        v_vars.append(iv['name'])
                else:
                    v_vars.append(iv)
            v['vars'] = v_vars
        # Get arguments
        self.arguments = []
        for k, v in self.inputs.items():
            self.arguments += v['vars']
        self.returns = []
        for k, v in self.outputs.items():
            self.returns += v['vars']
        self.debug("arguments: %s, returns: %s", self.arguments, self.returns)
        self.runner.pause()
        if service_address:
            os.remove(model_yaml)

    # def widget_function(self, *args, **kwargs):
    #     # import matplotlib.pyplot as plt
    #     # ncols = min(3, len(arguments))
    #     # nrows = int(ceil(float(len(arguments))/float(ncols)))
    #     # plt.show()
    #     out = self(*args, **kwargs)
    #     return out

    # def widget(self, *args, **kwargs):
    #     from ipywidgets import interact_manual
    #     return interact_manual(self.widget_function, *args, **kwargs)
        
    def __call__(self, *args, **kwargs):
        r"""Call the model as a function by sending variables.

        Args:
           *args: Any positional arguments are expected to be input variables
               in the correct order.
           **kwargs: Any keyword arguments are expected to be named input
               variables for the model.

        Raises:
            RuntimeError: If an input argument is missing.
            RuntimeError: If sending an input argument to a model fails.
            RuntimeError: If receiving an output value from a model fails.

        Returns:
            dict: Returned values for each return variable.

        """
        self.runner.resume()
        # Check for arguments
        for a, arg in zip(self.arguments, args):
            assert(a not in kwargs)
            kwargs[a] = arg
        for a in self.arguments:
            if a not in kwargs:  # pragma: debug
                raise RuntimeError("Required argument %s not provided." % a)
        # Send
        for k, v in self.inputs.items():
            flag = v['comm'].send([kwargs[a] for a in v['vars']])
            if not flag:  # pragma: debug
                raise RuntimeError("Failed to send %s" % k)
        # Receive
        out = {}
        for k, v in self.outputs.items():
            flag, data = v['comm'].recv(timeout=60.0)
            if not flag:  # pragma: debug
                raise RuntimeError("Failed to receive variable %s" % v)
            ivars = v['vars']
            if isinstance(data, (list, tuple)):
                assert(len(data) == len(ivars))
                for a, d in zip(ivars, data):
                    out[a] = d
            else:
                assert(len(ivars) == 1)
                out[ivars[0]] = data
        self.runner.pause()
        return out

    def stop(self):
        r"""Stop the model(s) from running."""
        self.runner.resume()
        if self._stop_called:
            return
        self._stop_called = True
        for x in self.inputs.values():
            x['comm'].send_eof()
        self.model_driver['instance'].set_break_flag()
        self.runner.waitModels(timeout=10)
        for x in self.inputs.values():
            x['comm'].close()
        for x in self.outputs.values():
            x['comm'].close()
        self.runner.terminate()
        self.runner.atexit()
        os.environ.clear()
        os.environ.update(self.old_environ)

    def model_info(self):
        r"""Display information about the wrapped model(s)."""
        print("Models: %s\nInputs:\n%s\nOutputs:\n%s\n"
              % (', '.join([x['name'] for x in
                            self.runner.modeldrivers.values()
                            if x['name'] != self.dummy_name]),
                 '\n'.join(['\t%s (vars=%s)' % (k, v['vars'])
                            for k, v in self.inputs.items()]),
                 '\n'.join(['\t%s (vars=%s)' % (k, v['vars'])
                            for k, v in self.outputs.items()])))


class YggRunner(YggClass):
    r"""This class handles the orchestration of starting the model and
    IO drivers, monitoring their progress, and cleaning up on exit.

    Arguments:
        modelYmls (list): List of paths to yaml files specifying the models
            that should be run.
        namespace (str, optional): Name that should be used to uniquely
            identify any RMQ exchange. Defaults to the value in the config
            file.
        host (str, optional): Name of the host that the models will be
            launched from. Defaults to None.
        rank (int, optional): Rank of this set of models if run in parallel.
            Defaults to 0.
        ygg_debug_level (str, optional): Level for Ygg debug messages.
            Defaults to environment variable 'YGG_DEBUG'.
        rmq_debug_level (str, optional): Level for RabbitMQ debug messages.
            Defaults to environment variable 'RMQ_DEBUG'.
        ygg_debug_prefix (str, optional): Prefix for Ygg debug messages.
            Defaults to namespace.
        as_service (bool, optional): If True, the integration is running as a
            service. If True, complete_partial is set to True. Defaults to
            False.
        complete_partial (bool, optional): If True, unpaired input/output
            channels are allowed and reserved for use (e.g. for calling the
            model as a function). Defaults to False.
        partial_commtype (dict, optional): Communicator kwargs that should be
            be used for the connections to the unpaired channels when
            complete_partial is True. Defaults to None and will be ignored.
        yaml_param (dict, optional): Parameters that should be used in
            mustache formatting of YAML files. Defaults to None and is
            ignored.
        validate (bool, optional): If True, the validation scripts for each
            modle (if present), will be run after the integration finishes
            running. Defaults to False.

    Attributes:
        namespace (str): Name that should be used to uniquely identify any
            RMQ exchange.
        host (str): Name of the host that the models will be launched from.
        rank (int): Rank of this set of models if run in parallel.
        modeldrivers (dict): Model drivers associated with this run.
        connectiondrivers (dict): Connection drivers for this run.
        interrupt_time (float): Time of last interrupt signal.
        error_flag (bool): True if one or more models raises an error.

    ..todo:: namespace, host, and rank do not seem strictly necessary.

    """
    def __init__(self, modelYmls, namespace=None, host=None, rank=0,
                 ygg_debug_level=None, rmq_debug_level=None,
                 ygg_debug_prefix=None, connection_task_method='thread',
                 as_service=False, complete_partial=False,
                 partial_commtype=None, production_run=False,
                 mpi_tag_start=None, yaml_param=None, validate=False):
        self.mpi_comm = None
        name = 'runner'
        if MPI is not None:
            comm = MPI.COMM_WORLD
            if comm.Get_size() > 1:
                self.mpi_comm = comm
                rank = comm.Get_rank()
                name += str(rank)
        super(YggRunner, self).__init__(name)
        if namespace is None:
            namespace = ygg_cfg.get('rmq', 'namespace', False)
        if not namespace:  # pragma: debug
            raise Exception('rmq:namespace not set in config file')
        if as_service:
            complete_partial = True
        self.namespace = namespace
        self.host = host
        self.rank = rank
        self.connection_task_method = connection_task_method
        self.base_dup = {}
        self.modelcopies = {}
        self.modeldrivers = {}
        self.connectiondrivers = {}
        self.interrupt_time = 0
        self._old_handlers = {}
        self.production_run = production_run
        self.error_flag = False
        self.complete_partial = complete_partial
        self.partial_commtype = partial_commtype
        self.validate = validate
        self.debug("Running in %s with path %s namespace %s rank %d",
                   os.getcwd(), sys.path, namespace, rank)
        # Update environment based on config
        cfg_environment()
        # Parse yamls
        self.mpi_tag_start = mpi_tag_start
        if self.mpi_comm and (self.rank > 0):
            pass
        else:
            self.drivers = yamlfile.parse_yaml(
                modelYmls, complete_partial=complete_partial,
                partial_commtype=partial_commtype, yaml_param=yaml_param)
            self.connectiondrivers = self.drivers['connection']
            self.modeldrivers = self.drivers['model']
            for x in self.modeldrivers.values():
                if x['driver'] == 'DummyModelDriver':
                    x['runner'] = self
                    if as_service:
                        for io in x['output_drivers']:
                            for comm in io['inputs']:
                                comm['for_service'] = True
                        for io in x['input_drivers']:
                            for comm in io['outputs']:
                                comm['for_service'] = True

    def pprint(self, *args):
        r"""Print with color."""
        s = ''.join(str(i) for i in args)
        print((COLOR_TRACE + '{}' + COLOR_NORMAL).format(s))

    def atexit(self, *args, **kwargs):
        r"""At exit ensure that the runner has stopped and cleaned up."""
        self.debug('')
        self.reset_signal_handler()
        self.closeChannels()
        self.cleanup()

    def signal_handler(self, sig, frame):
        r"""Terminate all drivers on interrrupt."""
        self.debug("Interrupt with signal %d", sig)
        now = time.perf_counter()
        elapsed = now - self.interrupt_time
        self.debug('Elapsed time since last interrupt: %d s', elapsed)
        self.interrupt_time = now
        self.pprint(' ')
        self.pprint(80 * '*')
        if elapsed < 5:
            self.pprint('* %76s *' % 'Interrupted twice within 5 seconds: shutting down')
            self.pprint(80 * '*')
            self.debug("Terminating models and closing all channels")
            self.terminate()
            self.pprint(80 * '*')
            return 1
        else:
            self.pprint('* %76s *' % 'Interrupted: Displaying channel summary')
            self.pprint('* %76s *' % 'interrupt again (within 5 seconds) to exit')
            self.pprint(80 * '*')
            self.printStatus()
            self.pprint(80 * '*')
        self.debug('%d returns', sig)

    def _swap_handler(self, signum, signal_handler):
        self._old_handlers[signum] = signal.getsignal(signum)
        signal.signal(signum, signal_handler)
        if not platform._is_win:
            signal.siginterrupt(signum, False)
        
    def set_signal_handler(self, signal_handler=None):
        r"""Set the signal handler.

        Args:
            signal_handler (function, optional): Function that should handle
                received SIGINT and SIGTERM signals. Defaults to
                self.signal_handler.

        """
        if signal_handler is None:
            signal_handler = self.signal_handler
        if signal_handler:
            self._swap_handler(signal.SIGINT, signal_handler)
            if not platform._is_win:
                self._swap_handler(signal.SIGTERM, signal_handler)
            else:  # pragma: windows
                self._swap_handler(signal.SIGBREAK, signal_handler)

    def reset_signal_handler(self):
        r"""Reset signal handlers to old ones."""
        for k, v in self._old_handlers.items():
            signal.signal(k, v)

    def run(self, signal_handler=None, timer=None, t0=None):
        r"""Run all of the models and wait for them to exit.

        Args:
            signal_handler (function, optional): Function that should be used as
                a signal handler. Defaults to None and is set by
                set_signal_handler.
            timer (function, optional): Function that should be called to get
                intermediate timing statistics. Defaults to time.time if not
                provided.
            t0 (float, optional): Zero point for timing statistics. Is set
                using the provided timer if not provided.

        Returns:
            dict: Intermediate times from the run.

        """
        with temp_config(production_run=self.production_run):
            if timer is None:
                timer = time.time
            if t0 is None:
                t0 = timer()
            times = OrderedDict()
            times['init'] = timer()
            self.loadDrivers()
            times['load drivers'] = timer()
            self.startDrivers()
            times['start drivers'] = timer()
            self.set_signal_handler(signal_handler)
            if not self.complete_partial:
                self.waitModels()
                times['run models'] = timer()
                self.atexit()
                times['at exit'] = timer()
            tprev = t0
            for k, t in times.items():
                self.info('%20s\t%f', k, t - tprev)
                tprev = t
            self.info(40 * '=')
            self.info('%20s\t%f', "Total", tprev - t0)
        if self.error_flag:
            raise IntegrationError("Error running the integration.")
        if self.validate:
            for v in self.modeldrivers.values():
                v['instance'].run_validation()
        return times

    @property
    def all_drivers(self):
        r"""iterator: For all drivers."""
        return chain(self.connectiondrivers.values(),
                     self.modeldrivers.values())

    def io_drivers(self):
        r"""Return the input and output drivers for all models.

        Returns:
            iterator: Access to list of I/O drivers.

        """
        return self.connectiondrivers.values()

    def create_driver(self, yml, **kwargs):
        r"""Create a driver instance from the yaml information.

        Args:
            yml (yaml): Yaml object containing driver information.

        Returns:
            object: An instance of the specified driver.

        """
        self.debug('Creating %s, a %s', yml['name'], yml['driver'])
        curpath = os.getcwd()
        if 'working_dir' in yml:
            os.chdir(yml['working_dir'])
        try:
            if (yml.get('copies', 1) > 1) and ('copy_index' not in yml):
                instance = DuplicatedModelDriver(
                    yml, namespace=self.namespace, rank=self.rank,
                    duplicates=yml.pop('duplicates', None), **kwargs)
            else:
                kwargs = dict(yml, **kwargs)
                instance = create_driver(yml=yml, namespace=self.namespace,
                                         rank=self.rank, **kwargs)
            yml['instance'] = instance
        finally:
            os.chdir(curpath)
        return instance

    def get_models(self, name, rank=None):
        r"""Get the set of drivers referenced by a model name.

        Args:
            name (str, list): Name of model(s).
            rank (int, optional): If provided, only models that will run on
                MPI processes with this rank will be returned. Defaults to
                None and is ignored.

        Returns:
            list: Set of drivers for a model.

        """
        if isinstance(name, list):
            models = []
            for x in name:
                models += self.get_models(x, rank=rank)
        elif name in self.modelcopies:
            models = [self.modeldrivers[cpy] for cpy in self.modelcopies[name]]
        elif name in self.modeldrivers:
            models = [self.modeldrivers[name]]
        else:
            models = [self.modeldrivers[
                DuplicatedModelDriver.get_base_name(name)]]
            assert(models[0].get('copies', 0) > 1)
        if rank is not None:  # pragma: debug
            # models = [x for x in models if (x['mpi_rank'] == rank)]
            raise NotImplementedError
        return models

    def bridge_mpi_connections(self, yml):
        r"""Bridge connections over MPI processes."""
        from yggdrasil.communication.MPIComm import MPIComm
        io_map = {'inputs': 'outputs', 'outputs': 'inputs'}
        models = {}
        for io in io_map.keys():
            models[io[:-1]] = [x['name'] for x in self.get_models(
                [x.get('partner_model', None) for x in yml[io]
                 if x.get('partner_model', None)])]
        for io, io_opp in io_map.items():
            for x in yml[io]:
                model = x.get('partner_model', None)
                if not model:
                    continue
                rank_map = {}
                for m in self.get_models(model):
                    rank_map.setdefault(m['mpi_rank'], [])
                    rank_map[m['mpi_rank']].append(m)
                if not any(rank > 0 for rank in rank_map.keys()):
                    continue
                if 'models' not in yml:
                    yml['models'] = models
                comms = []
                for rank in rank_map.keys():
                    x_copy = dict(copy.deepcopy(x),
                                  partner_copies=len(rank_map[rank]))
                    if rank == 0:
                        icomm = x_copy
                    else:
                        icomm = dict(
                            commtype='mpi',
                            daemon=True,
                            ranks=[rank],
                            mpi_index=len(self._mpi_comms),
                            mpi_direction=io_opp,
                            mpi_stride=1,
                            mpi_driver={
                                io_opp: [{'commtype': 'mpi', 'ranks': [0],
                                          'daemon': True}],
                                io: [x_copy],
                                'driver': yml['driver'],
                                'name': (
                                    '%s_mpi%s_%s' % (yml['name'], rank, io)),
                                'models': {
                                    io_opp[:-1]: models[io_opp[:-1]],
                                    io[:-1]: [m['name'] for m in
                                              rank_map[rank]]}})
                        if yml['driver'].startswith('RPC'):
                            icomm['mpi_stride'] += MPIComm._max_response
                        self._mpi_comms.append(icomm)
                        for m in rank_map[rank]:
                            drv_key = 'mpi_%s_drivers' % io_opp[:-1]
                            m.setdefault(drv_key, [])
                            m[drv_key].append(icomm['mpi_driver']['name'])
                    comms.append(icomm)
                if len(comms) == 1:
                    x.update(comms[0])
                    self._mpi_comms[comms[0]['mpi_index']] = x
                else:
                    # TODO: Move to connection level?
                    x.clear()
                    x['commtype'] = comms
                    if yml['driver'].startswith('RPC'):
                        x['pattern'] = 'cycle'

    def create_connection_driver(self, yml):
        r"""Create a connection driver instance from the yaml information.

        Args:
            yml (yaml): Yaml object containing driver information.

        Returns:
            object: An instance of the specified driver.

        """
        yml['task_method'] = self.connection_task_method
        drv = self.create_driver(yml)
        # Transfer connection addresses to model via env
        # TODO: Change to server that tracks connections
        for model, env in drv.model_env.items():
            env_key = 'env'
            if (model not in self.modelcopies) and (model not in self.modeldrivers):
                env_key = 'env_%s' % model
            for x in self.get_models(model):
                x.setdefault(env_key, {})
                x[env_key].update(env)
        return drv

    def distribute_mpi(self):
        r"""Distribute models between MPI processes."""
        size = self.mpi_comm.Get_size()
        if self.rank == 0:
            from yggdrasil.communication.MPIComm import MPIComm
            self.expand_duplicates()
            # Set the rank and index for each model
            for i, v in enumerate(self.modeldrivers.values()):
                v['mpi_rank'] = (i + 1) % size
                v['model_index'] = i
                v['mpi_tag_start'] = self.mpi_tag_start
            # Split the connections bridging MPI processes
            self.debug("Splitting connection drivers over MPI")
            self.all_connectiondrivers = self.connectiondrivers
            self._mpi_comms = []
            for driver in self.connectiondrivers.values():
                self.bridge_mpi_connections(driver)
            tag_start = len(ModelDriver._mpi_tags) * len(self.modeldrivers) * 5
            if self.mpi_tag_start is not None:
                tag_start += self.mpi_tag_start
            tag_stride = sum([x.pop('mpi_stride') for x in self._mpi_comms])
            connections = [[] for _ in range(size)]
            for x in self._mpi_comms:
                x['tag_start'] = tag_start + x.pop('mpi_index') * MPIComm._spacer_tags
                x['tag_stride'] = tag_stride
                io = x.pop('mpi_direction')
                drv = x.pop('mpi_driver')
                drv[io][0]['tag_start'] = x['tag_start']
                drv[io][0]['tag_stride'] = x['tag_stride']
                connections[x['ranks'][0]].append((drv['name'], drv))
            max_len = len(max(connections, key=len))
            for x in connections:
                while len(x) < max_len:
                    x.append(None)
            # Sort models
            self.all_modeldrivers = self.modeldrivers
            models = [[] for _ in range(size)]
            for i, (k, v) in enumerate(self.modeldrivers.items()):
                x_cp = copy.deepcopy(v)
                for k2 in ['input_drivers', 'output_drivers', 'mpi_rank']:
                    x_cp.pop(k2, None)
                for k2 in ['input_drivers', 'output_drivers']:
                    x_cp[k2] = x_cp.get('mpi_%s' % k2, [])
                # Skew models away from root process so that
                # connection threading might not share process
                models[v['mpi_rank']].append((k, x_cp))
            max_len = len(max(models, key=len))
            for x in models:
                while len(x) < max_len:
                    x.append(None)
        else:
            models = None
            connections = None
        self.modeldrivers = dict(
            [x for x in self.mpi_comm.scatter(models, root=0)
             if (x is not None)])
        self.connectiondrivers = dict(
            [x for x in self.mpi_comm.scatter(connections, root=0)
             if (x is not None)])
        self.modelcopies = self.mpi_comm.bcast(self.modelcopies, root=0)
        self.info("Models on MPI process %d: %s", self.rank,
                  list(self.modeldrivers.keys()))
        # Add dummy drivers on root process to monitor remote ones
        # and re-group copies into duplicate model w/ duplicate models
        # before non-duplicate to allow them to start before starting
        # local models
        if self.rank == 0:
            for i, (k, v) in enumerate(self.all_modeldrivers.items()):
                if k not in self.modeldrivers:
                    v['partner_driver'] = v['driver']
                    v['language'] = 'mpi'
                    v['driver'] = 'MPIPartnerModel'
                self.modeldrivers[k] = v
            self.connectiondrivers = self.all_connectiondrivers
        else:
            for v in self.modeldrivers.values():
                for k in ['input_drivers', 'output_drivers']:
                    v[k] = [self.connectiondrivers[x] for x in v.get(k, [])]
        self.reduce_duplicates()

    def expand_duplicates(self):
        r"""Expand model copies so they can be split across MPI processes."""
        self.debug("Expanding duplicated models")
        remove_dup = []
        add_dup = {}
        for k, v in self.modeldrivers.items():
            if v.get('copies', 1) > 1:
                self.modelcopies[v['name']] = []
                for x in DuplicatedModelDriver.get_yaml_copies(v):
                    add_dup[x['name']] = x
                    self.modelcopies[v['name']].append(x['name'])
                remove_dup.append(k)
        for k in remove_dup:
            self.base_dup[k] = self.modeldrivers.pop(k)
        self.modeldrivers.update(add_dup)

    def reduce_duplicates(self):
        r"""Join model duplicates after they were split between processes."""
        self.debug("Reducing duplicated models")
        for k in list(self.modelcopies.keys()):
            duplicates = [self.modeldrivers.pop(cpy)
                          for cpy in self.modelcopies.pop(k)
                          if cpy in self.modeldrivers]
            if duplicates:
                if k in self.base_dup:
                    base = self.base_dup[k]
                else:
                    base = dict(copy.deepcopy(duplicates[0]), name=k,
                                input_drivers=duplicates[0].get(
                                    'input_drivers', []),
                                output_drivers=duplicates[0].get(
                                    'output_drivers', []))
                    base.pop('copy_index', None)
                for x in duplicates:
                    for k2 in ['input_drivers', 'output_drivers']:
                        x[k2] = base.get(k2, [])
                base['duplicates'] = duplicates
                self.modeldrivers[k] = base
        
    def loadDrivers(self):
        r"""Load all of the necessary drivers, doing the IO drivers first
        and adding IO driver environmental variables back tot he models."""
        self.debug('')
        driver = dict(name='name')
        try:
            # Preparse model drivers first so that the input/output
            # channels are updated for wrapped functions
            self.debug("Preparsing model functions")
            for driver in self.modeldrivers.values():
                driver_cls = import_component('model', driver['driver'],
                                              without_schema=True)
                driver_cls.preparse_function(driver)
            if self.mpi_comm:
                self.distribute_mpi()
            # Create I/O drivers
            self.debug("Loading connection drivers")
            for driver in self.connectiondrivers.values():
                driver['task_method'] = self.connection_task_method
                self.create_connection_driver(driver)
            # Create model drivers
            self.debug("Loading model drivers")
            for driver in self.modeldrivers.values():
                self.create_driver(driver)
                self.debug("Model %s:, env: %s",
                           driver['name'], pformat(driver['instance'].env))
        except BaseException as e:  # pragma: debug
            self.error("%s could not be created: %s", driver['name'], e)
            self.terminate()
            raise

    def start_server(self, name):
        r"""Start a server driver."""
        if self.mpi_comm and (self.rank != 0):
            return
        # This is required if modelcopies are not joined before drivers
        # are started
        # if name in self.modelcopies:
        #     assert(name not in self.modeldrivers)
        #     for cpy in self.modelcopies[name]:
        #         self.start_server(cpy)
        #     return
        x = self.modeldrivers[name]['instance']
        if not x.was_started:
            self.debug("Starting server '%s' before client", x.name)
            x.start()

    def stop_server(self, name):
        r"""Stop a server driver."""
        # This is required if modelcopies are not joined before drivers
        # are started
        # if name in self.modelcopies:
        #     assert(name not in self.modeldrivers)
        #     for cpy in self.modelcopies[name]:
        #         self.stop_server(cpy)
        #     return
        x = self.modeldrivers[name]['instance']
        x.stop()

    def startDrivers(self):
        r"""Start drivers, starting with the IO drivers."""
        if not self.mpi_comm or (self.rank == 0):
            assert(not self.modelcopies)
        self.info('Starting I/O drivers and models on system '
                  + '{} in namespace {} with rank {}'.format(
                      self.host, self.namespace, self.rank))
        driver = dict(name='name')
        try:
            # Start connections
            for driver in self.io_drivers():
                self.debug("Starting driver %s", driver['name'])
                d = driver['instance']
                if not d.was_started:
                    d.start()
            # Ensure connections in loop
            for driver in self.io_drivers():
                self.debug("Checking driver %s", driver['name'])
                d = driver['instance']
                d.wait_for_loop()
                assert(d.was_loop)
                assert(not d.errors)
            # Start models
            for driver in self.modeldrivers.values():
                self.debug("Starting driver %s", driver['name'])
                d = driver['instance']
                for n2 in driver.get('client_of', []):
                    self.start_server(n2)
                if not d.was_started:
                    d.start()
        except BaseException as e:  # pragma: debug
            self.error("%s did not start: %s(%s)", driver['name'], type(e), e)
            self.terminate()
            raise
        if self.mpi_comm:
            self.mpi_comm.barrier()
        self.debug('ALL DRIVERS STARTED')

    @property
    def is_alive(self):
        r"""bool: True if all of the models are still running, False
        otherwise."""
        for drv in self.modeldrivers.values():
            if (not drv['instance'].is_alive()) or drv['instance'].errors:
                return False
        return True

    def waitModels(self, timeout=False):
        r"""Wait for all model drivers to finish. When a model finishes,
        join the thread and perform exits for associated IO drivers."""
        self.debug('')
        running = [d for d in self.modeldrivers.values()]
        dead = []
        Tout = self.start_timeout(t=timeout,
                                  key_suffix='.waitModels')
        while ((len(running) > 0) and (not self.error_flag)
               and (not Tout.is_out)):
            for drv in running:
                d = drv['instance']
                if d.errors:  # pragma: debug
                    self.error('Error in model %s', drv['name'])
                    self.error_flag = True
                    break
                elif d.io_errors:  # pragma: debug
                    self.error('Error in input/output driver for model %s'
                               % drv['name'])
                    self.error_flag = True
                    break
                d.join(1)
                if not d.is_alive():
                    if not d.errors:
                        self.info("%s finished running.", drv['name'])
                        # self.do_model_exits(drv)
                        # self.debug("%s completed model exits.", drv['name'])
                        self.do_client_exits(drv)
                        self.debug("%s completed client exits.", drv['name'])
                        running.remove(drv)
                        self.info("%s finished exiting.", drv['name'])
                else:
                    self.debug('%s still running', drv['name'])
            dead = []
            for drv in self.all_drivers:
                d = drv['instance']
                d.join(0.1)
                if not d.is_alive():
                    dead.append(drv['name'])
        self.stop_timeout(key_suffix='.waitModels')
        for d in self.modeldrivers.values():
            if d['instance'].errors:
                self.error_flag = True
        if not self.error_flag:
            self.info('All models completed')
        else:
            self.error('One or more models generated errors.')
            self.printStatus()
            self.terminate()
        if self.mpi_comm:
            allcode = self.mpi_comm.allreduce(self.error_flag, op=MPI.SUM)
            if not self.error_flag:
                self.error_flag = allcode
        self.debug('Returning')

    # def do_model_exits(self, model):
    #     r"""Perform exits for IO drivers associated with a model.

    #     Args:
    #         model (dict): Dictionary of model parameters including any
    #             associated IO drivers.

    #     """
    #     for drv in model['input_drivers']:
    #         #  if model['name'] in drv['models']:
    #         #     drv['models'].remove(model['name'])
    #         if not drv['instance'].is_alive():
    #             continue
    #         # if (len(drv['models']) == 0):
    #         self.debug('on_model_exit %s', drv['name'])
    #         drv['instance'].on_model_exit('output', model['name'])
    #     for drv in model['output_drivers']:
    #         # if model['name'] in drv['models']:
    #         #     drv['models'].remove(model['name'])
    #         if not drv['instance'].is_alive():
    #             continue
    #         # if (len(drv['models']) == 0):
    #         self.debug('on_model_exit %s', drv['name'])
    #         drv['instance'].on_model_exit('input', model['name'])
    
    def do_client_exits(self, model):
        r"""Perform exits for IO drivers associated with a client model.

        Args:
            model (dict): Dictionary of model parameters including any
                associated IO drivers.

        """
        if self.mpi_comm and (self.rank != 0):
            return
        # TODO: Exit upstream models that no longer have any open
        # output, connections when a connection is closed.
        for srv_name in model.get('client_of', []):
            iod = self.connectiondrivers[srv_name]
            iod['instance'].remove_model('input', model['name'])
            if iod['instance'].nclients == 0:
                self.stop_server(srv_name)

    def pause(self):
        r"""Pause all drivers."""
        self.debug('')
        for driver in self.all_drivers:
            if 'instance' in driver:
                driver['instance'].pause()

    def resume(self):
        r"""Resume all paused drivers."""
        self.debug('')
        for driver in self.all_drivers:
            if 'instance' in driver:
                driver['instance'].resume()

    def terminate(self):
        r"""Immediately stop all drivers, beginning with IO drivers."""
        self.debug('')
        self.resume()
        for driver in self.all_drivers:
            if 'instance' in driver:
                self.debug('Stop %s', driver['name'])
                driver['instance'].terminate()
                # Terminate should ensure instance not alive
                assert(not driver['instance'].is_alive())
        self.debug('Returning')

    def cleanup(self):
        r"""Perform cleanup operations for all drivers."""
        self.debug('')
        for driver in self.all_drivers:
            if 'instance' in driver:
                driver['instance'].cleanup()
        # self.inputdrivers = {}
        # self.outputdrivers = {}
        # self.modeldrivers = {}

    def printStatus(self, return_str=False):
        r"""Print the status of all drivers, starting with the IO drivers."""
        self.debug('')
        out = []
        for driver in self.all_drivers:
            if 'instance' in driver:
                out.append(
                    driver['instance'].printStatus(return_str=return_str))
        if return_str:
            return '\n'.join(out)

    def closeChannels(self, force_stop=False):
        r"""Stop IO drivers and join the threads.

        Args:
            force_stop (bool, optional): If True, the terminate method is
                used to stop the drivers. Otherwise, the stop method is used.
                The stop method will try to exit gracefully while terminate
                will exit as quickly as possible. Defaults to False.

        """
        self.debug('')
        drivers = [i for i in self.io_drivers()]
        for drv in drivers:
            if 'instance' in drv:
                driver = drv['instance']
                if driver.is_alive():  # pragma: debug
                    self.debug("Stopping %s", drv['name'])
                    if force_stop or self.error_flag:
                        driver.terminate()
                    else:
                        driver.stop()
                    self.debug("Stop(%s) returns", drv['name'])
        self.debug('Channel Stops DONE')
        for drv in drivers:
            if 'instance' in drv:
                driver = drv['instance']
                assert(not driver.is_alive())
        self.debug('Returning')

        
def get_runner(models, **kwargs):
    r"""Get runner for a set of models, getting run information from the
    environment.

    Args:
        models (list): List of yaml files containing information on the models
            that should be run.
        **kwargs: Additonal keyword arguments are passed to YggRunner.

    Returns:
        YggRunner: Runner for the provided models.

    Raises:
        Exception: If config option 'namespace' in 'rmq' section not set.

    """
    # Get environment variables
    rank = os.environ.get('PARALLEL_SEQ', '0')
    host = socket.gethostname()
    os.environ['YGG_RANK'] = rank
    os.environ['YGG_HOST'] = host
    rank = int(rank)
    kwargs.update(rank=rank, host=host)
    # Run
    yggRunner = YggRunner(models, **kwargs)
    return yggRunner


def run(*args, **kwargs):
    yggRunner = get_runner(*args, **kwargs)
    yggRunner.run()
    yggRunner.debug("runner returns, exiting")



import pytest
import numpy as np
from yggdrasil import units
from yggdrasil.metaschema import MetaschemaTypeError
from tests.metaschema.properties.test_MetaschemaProperty import (
    TestMetaschemaProperty as base_class)


class TestSubtypeMetaschemaProperty(base_class):
    r"""Test class for SubtypeMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.ScalarMetaschemaProperties'
    _cls = 'SubtypeMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def valid(self):
        r"""Objects that are valid."""
        return [(int(1), 'int'), (float(1), 'float')]

    @pytest.fixture(scope="class")
    def invalid(self):
        r"""Objects that are invalid."""
        return [(int(1), 'float'), (float(1), 'int')]

    @pytest.fixture(scope="class")
    def valid_compare(self):
        r"""Objects that successfully compare."""
        return [('int', 'int')]

    @pytest.fixture(scope="class")
    def invalid_compare(self):
        r"""Objects that do not successfully compare."""
        return [('float', 'int')]

    @pytest.fixture(scope="class")
    def valid_normalize_schema(self):
        r"""Schemas for normalization."""
        return [
            ({'subtype': 'float'}, {'subtype': 'float'}),
            ({'units': 'g'}, {'units': 'g', 'subtype': 'float'}),
            ({'units': ''}, {'units': ''})]

    def test_invalid_encode(self, python_class):
        r"""Test invalid encode for object dtype."""
        with pytest.raises(MetaschemaTypeError):
            python_class.encode(object)


class TestPrecisionMetaschemaProperty(base_class):
    r"""Test class for PrecisionMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.ScalarMetaschemaProperties'
    _cls = 'PrecisionMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def valid(self):
        r"""Objects that are valid."""
        return [(np.int32(1), 32), (np.float16(1), 16)]

    @pytest.fixture(scope="class")
    def invalid(self):
        r"""Objects that are invalid."""
        return [(np.int32(1), 8), (np.float32(1), 16)]

    @pytest.fixture(scope="class")
    def valid_compare(self):
        r"""Objects that successfully compare."""
        return [(32, 32), (16, 32)]

    @pytest.fixture(scope="class")
    def invalid_compare(self):
        r"""Objects that do not successfully compare."""
        return [(32, 16)]

    @pytest.fixture(scope="class")
    def valid_normalize_schema(self):
        r"""Schemas for normalization."""
        return [
            ({'precision': 64}, {'precision': 64}),
            ({'subtype': 'int'}, {'subtype': 'int', 'precision': 64}),
            ({'subtype': 'complex'}, {'subtype': 'complex', 'precision': 128})]


class TestUnitsMetaschemaProperty(base_class):
    r"""Test class for UnitsMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.ScalarMetaschemaProperties'
    _cls = 'UnitsMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def valid(self):
        r"""Objects that are valid."""
        return [(1, ''), (units.add_units(1, 'cm'), 'm')]

    @pytest.fixture(scope="class")
    def invalid(self):
        r"""Objects that are invalid."""
        return [(units.add_units(1, 'cm'), 'kg')]

    @pytest.fixture(scope="class")
    def valid_compare(self):
        r"""Objects that successfully compare."""
        return [('cm', 'cm'), ('cm', 'm'), ('m', 'cm'), ('', 'cm'), ('cm', '')]

    @pytest.fixture(scope="class")
    def invalid_compare(self):
        r"""Objects that do not successfully compare."""
        return [('cm', 'g')]



import os
import glob
import copy
import six
import inspect
import importlib
import contextlib
import weakref
from collections import OrderedDict
from yggdrasil.doctools import docs2args


_registry = {}
_registry_complete = False


class ComponentError(BaseException):
    r"""Error raised when there is a problem import a component."""
    pass


class ClassRegistry(OrderedDict):
    r"""Class for registering classes."""

    def __init__(self, *args, import_function=None, **kwargs):
        module = inspect.getmodule(inspect.stack()[1][0])
        self._module = module.__name__
        self._directory = os.path.dirname(module.__file__)
        self._import_function = import_function
        self._imported = False
        super(ClassRegistry, self).__init__(*args, **kwargs)

    def import_classes(self):
        r"""Import all classes in the same directory."""
        if self._imported:
            return
        self._imported = True
        for x in sorted(glob.glob(os.path.join(self._directory, '*.py'))):
            mod = os.path.basename(x)[:-3]
            if not mod.startswith('__'):
                importlib.import_module(self._module + '.%s' % mod)
        if self._import_function is not None:
            self._import_function()

    def keys(self, *args, **kwargs):
        self.import_classes()
        return super(ClassRegistry, self).keys(*args, **kwargs)

    def values(self, *args, **kwargs):
        self.import_classes()
        return super(ClassRegistry, self).values(*args, **kwargs)

    def items(self, *args, **kwargs):
        self.import_classes()
        return super(ClassRegistry, self).items(*args, **kwargs)

    def __contains__(self, key):
        self.import_classes()
        return super(ClassRegistry, self).__contains__(key)

    def get(self, key, default=None):
        if (not self.has_entry(key)):
            self.import_classes()
        return super(ClassRegistry, self).get(key, default)

    def __getitem__(self, *args, **kwargs):
        try:
            return super(ClassRegistry, self).__getitem__(*args, **kwargs)
        except KeyError:  # pragma: no cover
            # This will only be called during import
            if self._imported:
                raise
            self.import_classes()
            return super(ClassRegistry, self).__getitem__(*args, **kwargs)

    def has_entry(self, key):
        return super(ClassRegistry, self).__contains__(key)


def registration_in_progress():
    r"""Determine if a registration is in progress."""
    return bool(os.environ.get('YGGDRASIL_REGISTRATION_IN_PROGRESS', None))


@contextlib.contextmanager
def registering(recurse=False):
    r"""Context for preforming registration."""
    if not recurse:
        assert(not registration_in_progress())
    try:
        previous = os.environ.get('YGGDRASIL_REGISTRATION_IN_PROGRESS', None)
        os.environ['YGGDRASIL_REGISTRATION_IN_PROGRESS'] = '1'
        yield
    finally:
        if previous is None:
            if 'YGGDRASIL_REGISTRATION_IN_PROGRESS' in os.environ:
                del os.environ['YGGDRASIL_REGISTRATION_IN_PROGRESS']
        else:
            os.environ['YGGDRASIL_REGISTRATION_IN_PROGRESS'] = previous


def init_registry(recurse=False):
    r"""Initialize the registries and schema."""
    from yggdrasil.tools import import_all_modules
    global _registry
    global _registry_complete
    with registering(recurse=recurse):
        import_all_modules(exclude=['yggdrasil.examples',
                                    'yggdrasil.languages',
                                    'yggdrasil.interface',
                                    'yggdrasil.timing'],
                           do_first=['yggdrasil.serialize'])
        _registry_complete = True
    return _registry


def get_registry(comptype=None):
    r"""Get the registry that should be used for looking up components.

    Args:
        comptype (str, optional): The name of a component to get the
            registry for. Defaults to None and the entire registry will be
            returned.

    """
    global _registry
    if registration_in_progress():
        out = _registry
    else:
        from yggdrasil import constants
        out = constants.COMPONENT_REGISTRY
    if comptype:
        if comptype not in out:  # pragma: debug
            raise Exception(f"Importing a component type that has not yet "
                            f"been registered: {comptype}")
        out = out[comptype]
    return out


def suspend_registry():
    r"""Suspend the registry by storing the global registries in a dictionary."""
    global _registry
    global _registry_complete
    out = {'_registry': _registry, '_registry_complete': _registry_complete}
    _registry = {}
    _registry_complete = False
    return out


def restore_registry(reg_dict):
    r"""Restore the registry to values in the provided dictionary."""
    global _registry
    global _registry_complete
    _registry = reg_dict['_registry']
    _registry_complete = reg_dict['_registry_complete']


def import_component(comptype, subtype=None, **kwargs):
    r"""Dynamically import a component by name.

    Args:
        comptype (str): Component type.
        subtype (str, optional): Component subtype. If subtype is not one of
            the registered subtypes for the specified comptype, subtype is
            treated as the name of class. Defaults to None if not provided and
            the default subtype defined in the schema for the specified
            component will be used.
        **kwargs: Additional keyword arguments are used to determine the
            subtype if it is None.

    Returns:
        class: Component class.

    Raises:
        ComponentError: If comptype is not a registered component type.
        ComponentError: If subtype is not a registered subtype or the name of
            a registered subtype class for the specified comptype.
            

    """
    registry = get_registry(comptype=comptype)
    if subtype is None:
        subtype = kwargs.get(registry["key"], None)
    if (comptype == 'comm') and (subtype is None):
        subtype = 'DefaultComm'
    if subtype is None:
        subtype = registry["default"]
    if subtype in registry["subtypes"]:
        class_name = registry["subtypes"][subtype]
    else:
        class_name = subtype
    # Check registered components to prevent importing multiple times
    if class_name not in registry.get("classes", {}):
        registry.setdefault("classes", {})
        try:
            registry["classes"][class_name] = getattr(
                importlib.import_module(f"{registry['module']}.{class_name}"),
                class_name)
        except ImportError:
            if comptype == 'comm':
                try:
                    return import_component('file', subtype, **kwargs)
                except ComponentError:
                    pass
            raise ComponentError(f"Could not locate a {comptype} component "
                                 f"{subtype}.")
    out_cls = registry["classes"][class_name]
    # Check for an aliased class
    if hasattr(out_cls, '_get_alias'):
        out_cls = out_cls._get_alias()
    return out_cls


def create_component(comptype, subtype=None, **kwargs):
    r"""Dynamically create an instance of a component with the specified options
    as outlined in the component schema. This function requires loading the
    component schemas and so should not be used at the module or class level to
    prevent circular dependencies.

    Args:
        comptype (str): Component type.
        subtype (str, optional): Component subtype. If subtype is not one of the
            registered subtypes for the specified comptype, subtype is treated
            as the name of the class. Defaults to None if not provided and the
            default subtype defined in the schema for the specified component
            will be used. If the subtype is specified by the component subtype
            key in the remaining kwargs, that subtype will be used instead.
        **kwargs: Additional keyword arguments are treated as options for the
            component as outlined in the component schema.

    Returns:
        ComponentBase: Instance of the specified component type/subtype and
            options.
    
    Raises:
        ComponentError: If comptype is not a registered component type.

    """
    from yggdrasil.schema import get_schema
    s = get_schema().get(comptype, None)
    if s is None:  # pragma: debug
        raise ComponentError("Unrecognized component type: %s" % comptype)
    if s.subtype_key in kwargs:
        subtype = kwargs[s.subtype_key]
    if subtype is None:
        subtype = s.identify_subtype(kwargs)
    cls = import_component(comptype, subtype=subtype, **kwargs)
    return cls(**kwargs)


def get_component_base_class(comptype, subtype=None, **kwargs):
    r"""Determine the base class for a component type.

    Args:
        comptype (str): The name of a component to test against.
        subtype (str, optional): Subtype to use to determine the component
            base class. Defaults to None.
        **kwargs: Additional keyword arguments are used to determine the
            subtype if it is None.
    
    Returns:
        ComponentBase: Component base class.

    """
    registry = get_registry(comptype=comptype)
    base_class_name = registry['base']
    return import_component(comptype, subtype=base_class_name, **kwargs)


def isinstance_component(x, comptype, subtype=None, **kwargs):
    r"""Determine if an object is an instance of a component type.

    Args:
        x (object): Object to test.
        comptype (str, list): The name of one or more components to test
            against.
        subtype (str, optional): Subtype to use to determine the component
            base class. Defaults to None.
        **kwargs: Additional keyword arguments are used to determine the
            subtype if it is None.

    Returns:
        bool: True if the object is an instance of the specified component(s).

    """
    if isinstance(comptype, (list, tuple)):
        for icomp in comptype:
            if isinstance_component(x, icomp):
                return True
        else:
            return False
    base_class = get_component_base_class(comptype, subtype=subtype,
                                          **kwargs)
    return isinstance(x, base_class)


def inherit_schema(orig, new_properties=None, new_required=None,
                   remove_keys=[], **kwargs):
    r"""Create an inherited schema, adding new value to accepted ones for
    dependencies.
    
    Args:
        orig (dict): Schema that will be inherited.
        new_properties (dict, optional): Dictionary of new properties to add.
            Defaults to None and is ignored.
        new_requried (list, optional): Properties that should be required by the
            new class. Defaults to None and is ignored.
        remove_keys (list, optional): Keys that should be removed form orig
            before adding the new keys. Defaults to empty list.
        **kwargs: Additional keyword arguments will be added to the schema
            with dependency on the provided key/value pair.

    Returns:
        tuple(dict, list): New schema properties and a list of requried.

    """
    remove_keys = copy.deepcopy(remove_keys)
    # Get set of original properties
    assert(issubclass(orig, ComponentBase))
    out_prp = copy.deepcopy(orig._schema_properties)
    if orig._schema_excluded_from_inherit is not None:
        remove_keys += orig._schema_excluded_from_inherit
    out_req = copy.deepcopy(orig._schema_required)
    # Don't add duplicates
    if new_properties == out_prp:
        new_properties = None
    if new_required == out_req:
        new_required = None
    # Remove keys
    for k in remove_keys:
        if k in out_prp:
            out_prp.pop(k)
        if k in out_req:
            out_req.remove(k)
    # Add new values and keyword arguments
    if new_properties is not None:
        out_prp.update(new_properties)
    if new_required is not None:
        for k in new_required:
            if k not in out_req:
                out_req.append(k)
    out_prp.update(kwargs)
    return out_prp, out_req


class ComponentMeta(type):
    r"""Meta class for registering schema components."""
    
    def __new__(meta, name, bases, class_dict):
        cls = type.__new__(meta, name, bases, class_dict)
        # Determine subtype
        subtype = None
        if cls._schema_subtype_key is not None:
            subtype = getattr(cls, cls._schema_subtype_key,
                              getattr(cls, '_' + cls._schema_subtype_key, None))
        # Inherit new schema properties
        if cls._schema_inherit and (cls.__name__ != 'ComponentBase'):
            assert(isinstance(cls._schema_inherit, bool))
            inherit_from = None
            for x in bases:
                if hasattr(x, '_schema_properties'):
                    inherit_from = x
                    break
            if inherit_from is None:  # pragma: debug
                raise RuntimeError(("Class %s dosn't have a component "
                                    "parent class.") % cls)
            # Dont inherit if the base is ComponentBase (empty schema)
            if inherit_from.__name__ != ComponentBase:
                cls._schema_properties, cls._schema_required = inherit_schema(
                    inherit_from,
                    new_properties=cls._schema_properties,
                    new_required=cls._schema_required,
                    remove_keys=cls._schema_excluded_from_inherit)
        # Do non-registration things
        if not (name.endswith('Base') or (subtype is None) or cls._dont_register):
            # Perform class specific actions in preparation for registration
            cls.before_registration(cls)
            # assert(cls.name is not None)
            # Add parameter descriptions from docs
            for x in cls.__mro__[::-1]:
                args_dict = docs2args(x.__doc__)
                for k, v in cls._schema_properties.items():
                    if k in args_dict:
                        v.setdefault('description',
                                     args_dict[k]['description'])
            # Determine base class
            if cls._schema_base_class is None:
                reg = get_registry()
                if cls._schema_type in reg:
                    cls._schema_base_class = reg[cls._schema_type]['base']
                else:
                    base_comp = cls.__name__
                    for i, x in enumerate(cls.__mro__):
                        if x._schema_type != cls._schema_type:
                            break
                        base_comp = x.__name__
                    else:  # pragma: debug
                        raise RuntimeError(
                            f"Could not determine base class for {cls} "
                            f"from {bases}.")
                    cls._schema_base_class = base_comp
            # Register
            global _registry
            yaml_typ = cls._schema_type
            default_subtype = cls._schema_properties.get(
                cls._schema_subtype_key, {}).get('default',
                                                 cls._schema_subtype_default)
            if yaml_typ not in _registry:
                _registry[yaml_typ] = OrderedDict([
                    ("classes", OrderedDict()),
                    ("module", '.'.join(cls.__module__.split('.')[:-1])),
                    ("default", default_subtype),
                    ("base", cls._schema_base_class),
                    ("key", cls._schema_subtype_key),
                    ("subtypes", {})])
            elif default_subtype is not None:
                assert(_registry[yaml_typ]["default"] == default_subtype)
            if cls.__name__ not in _registry[yaml_typ]["classes"]:
                _registry[yaml_typ]["classes"][cls.__name__] = cls
                _registry[yaml_typ]["subtypes"][subtype] = cls.__name__
            if not registration_in_progress():
                cls.after_registration(cls)
                cls.finalize_registration(cls)
        return cls


class ComponentBaseUnregistered(object):
    r"""Base class for schema components w/o schema and registration."""
    __slots__ = []
    _disconnect_attr = []

    def __del__(self):
        self.disconnect()

    def disconnect(self):
        r"""Disconnect attributes that are aliases."""
        for k in self._disconnect_attr:
            if hasattr(getattr(self, k, None), 'disconnect'):
                getattr(self, k).disconnect()


@six.add_metaclass(ComponentMeta)
class ComponentBase(ComponentBaseUnregistered):
    r"""Base class for schema components.

    Args:
        skip_component_schema_normalization (bool, optional): If True, the
            schema will not be used to normalize/validate input keyword
            arguments (e.g. in case they were already parsed). Defaults to
            False.
        **kwargs: Keyword arguments are added to the class as attributes
            according to the class attributes _schema_properties and
            _schema_excluded_from_class. Keyword arguments not added to the
            class as attributes are assigned to the extra_kwargs dictionary.

    Attributes:
        extra_kwargs (dict): Keyword arguments that were not parsed.

    Class Attributes:
        _schema_type (str): Name of the component type the class represents.
        _schema_subtype_key (str): Attribute that should be used to identify the
            subtype associated with each class.
        _schema_subtype_description (str): Description for the subtype represented
            by the class that should be used in documentation tables.
        _schema_required (list): Keys from _schema_properties that are required
            to produce a valid component.
        _schema_properties (dict): Schemas describing keyword arguments that
            can be supplied to the class constructor and used to specify
            component behavior in YAML/JSON files. At initialization, these
            keywords are added to the class instance as attributes of the
            same name unless they are in _schema_excluded_from_class. Unless
            _schema_inherit is False, these properties will be added in addition
            to the schema properties defined by the class base.
        _schema_excluded_from_class (list): Keywords in _schema_properties that
            should not be added to the class as attributes during initialization.
        _schema_excluded_from_inherit (list): Keywords in _schema_properties that
            should not be inherited either from the base class or by child
            classes.
        _schema_inherit (bool, ComponentBase): If False, the base schema will
            not be inherited. If a Component subclass, the schema from that
            class will be inherited instead of the base class. Defaults to True.
        _dont_register (bool): If True, the component class will be be registered
            and the before_registration class method will not be called. Defaults
            to False.
        

    """

    _schema_type = None
    _schema_subtype_key = None
    _schema_subtype_description = None
    _schema_subtype_default = None
    _schema_base_class = None
    _schema_required = []
    _schema_properties = {}
    _schema_excluded_from_class = []
    _schema_excluded_from_inherit = []
    _schema_excluded_from_class_validation = []
    _schema_inherit = True
    _dont_register = False

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj._input_args = []
        for x in args:
            try:
                obj._input_args.append(weakref.ref(x))
            except TypeError:
                obj._input_args.append(x)
        obj._input_kwargs = {}
        for k, v in kwargs.items():
            try:
                obj._input_kwargs[k] = weakref.ref(v)
            except TypeError:
                obj._input_kwargs[k] = v
        return obj
    
    def __getstate__(self):
        out = self.__dict__.copy()
        del out['_input_args'], out['_input_kwargs']
        return out

    def __setstate__(self, state):
        state['_input_args'] = []
        state['_input_kwargs'] = {}
        self.__dict__.update(state)

    def __init__(self, skip_component_schema_normalization=None, **kwargs):
        if skip_component_schema_normalization is None:
            skip_component_schema_normalization = (
                not (os.environ.get('YGG_VALIDATE_COMPONENTS', 'None').lower()
                     in ['true', '1']))
        comptype = self._schema_type
        if (comptype is None) and (not self._schema_properties):
            self.extra_kwargs = kwargs
            return
        subtype = None
        if self._schema_subtype_key is not None:
            subtype = getattr(self, self._schema_subtype_key,
                              getattr(self, '_%s' % self._schema_subtype_key, None))
        # Fall back to some simple parsing/normalization to save time on
        # full jsonschema normalization
        self._defaults_set = []
        for k, v in self._schema_properties.items():
            if k in self._schema_excluded_from_class:
                continue
            default = v.get('default', None)
            if (k == self._schema_subtype_key) and (subtype is not None):
                default = subtype
            if default is not None:
                if k not in kwargs:
                    self._defaults_set.append(k)
                kwargs.setdefault(k, copy.deepcopy(default))
            if v.get('type', None) == 'array':
                if isinstance(kwargs.get(k, None), (bytes, str)):
                    kwargs[k] = kwargs[k].split()
        # Parse keyword arguments using schema
        if (((comptype is not None) and (subtype is not None)
             and (not skip_component_schema_normalization)
             and (not self._dont_register))):
            from yggdrasil.schema import get_schema
            s = get_schema().get_component_schema(
                comptype, subtype, relaxed=True,
                allow_instance_definitions=True)
            props = list(s['properties'].keys())
            if not skip_component_schema_normalization:
                from yggdrasil import metaschema
                kwargs.setdefault(self._schema_subtype_key, subtype)
                # Remove properties that shouldn't ve validated in class
                for k in self._schema_excluded_from_class_validation:
                    if k in s['properties']:
                        del s['properties'][k]
                # Validate and normalize
                metaschema.validate_instance(kwargs, s, normalize=False)
                # TODO: Normalization performance needs improvement
                # import pprint
                # print('before')
                # pprint.pprint(kwargs_comp)
                # kwargs_comp = metaschema.validate_instance(kwargs_comp, s,
                #                                            normalize=True)
                # kwargs.update(kwargs_comp)
                # print('normalized')
                # pprint.pprint(kwargs_comp)
        else:
            props = self._schema_properties.keys()
        # Set attributes based on properties
        for k in props:
            if k in self._schema_excluded_from_class:
                continue
            v = kwargs.pop(k, None)
            if getattr(self, k, None) is None:
                setattr(self, k, v)
            # elif (getattr(self, k) != v) and (v is not None):
            #     warnings.warn(("The schema property '%s' is provided as a "
            #                    "keyword with a value of %s, but the class "
            #                    "already has an attribute of the same name "
            #                    "with the value %s.")
            #                   % (k, v, getattr(self, k)))
        self.extra_kwargs = kwargs

    @staticmethod
    def before_registration(cls):
        r"""Operations that should be performed to modify class attributes prior
        to registration. These actions will still be performed if the environment
        variable YGGDRASIL_REGISTRATION_IN_PROGRESS is set."""
        pass

    @staticmethod
    def after_registration(cls):
        r"""Operations that should be preformed to modify class attributes after
        registration. These actions will not be performed if the environment
        variable YGGDRASIL_REGISTRATION_IN_PROGRESS is set."""
        pass

    @staticmethod
    def finalize_registration(cls):
        r"""Final operations to perform after a class has been fully initialized.
        These actions will not be performed if the environment variable
        YGGDRASIL_REGISTRATION_IN_PROGRESS is set."""
        pass



import os


def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        output_dir=None,
        input_files=['input.txt'],
        output_files=['output.txt'],
        expected_output_files=[os.path.join('Output', 'expected_output.txt')])



import os
import sys
import signal
import uuid
import json
import traceback
import yaml
import glob
import pprint
import functools
import threading
import logging
from yggdrasil import runner
from yggdrasil import platform
from yggdrasil.multitasking import wait_on_function, ValueEvent
from yggdrasil.tools import YggClass, kill
from yggdrasil.config import ygg_cfg


_service_host_env = 'YGGDRASIL_SERVICE_HOST_URL'
_service_repo_dir = 'YGGDRASIL_SERVICE_REPO_DIR'
_default_service_type = ygg_cfg.get('services', 'default_type', 'flask')
_default_commtype = ygg_cfg.get('services', 'default_comm', None)
_default_address = ygg_cfg.get('services', 'address', None)
_client_id = ygg_cfg.get('services', 'client_id', None)
if platform._is_win:  # pragma: windows
    _shutdown_signal = signal.SIGBREAK
else:
    _shutdown_signal = signal.SIGTERM


class ClientError(BaseException):
    r"""Error raised by errors when calling the server from the client."""
    pass


class ServerError(BaseException):
    r"""Error raised when there was an error on the server."""
    pass


class ServiceBase(YggClass):
    r"""Base class for sending/responding to service requests.

    Args:
        name (str): Name that should be used to initialize an address for the
            service.
        for_request (bool, optional): If True, a client-side connection is
            initialized. If False a server-side connection is initialized.
            Defaults to False.
        address (str, optional): The address that the service can be accessed
            from. Defaults to ('services', 'address') configuration option, if
            set, and None if not.
        *args: Additional arguments are used to initialize the client/server
            connection.
        **kwargs: Additional keyword arguments are used to initialize the
            client/server connection.

    """

    default_address = None
    default_port = None

    def __init__(self, name, *args, **kwargs):
        self.for_request = kwargs.pop('for_request', False)
        self.address = kwargs.pop('address', None)
        self.port = kwargs.pop('port', None)
        if self.address is None:
            self.address = _default_address
        if self.address is None:
            self.address = self.default_address
        if self.port is None:
            self.port = self.default_port
        if isinstance(self.address, str) and ('{port}' in self.address):
            self.address = self.address.format(port=self.port)
        self._args = args
        self._kwargs = kwargs
        self._is_running = False
        super(ServiceBase, self).__init__(name, *args, **kwargs)
        if self.for_request:
            self.setup_client(*args, **kwargs)
        else:
            self.setup_server(*args, **kwargs)

    @classmethod
    def is_installed(cls):
        r"""bool: True if the class is fully installed, False otherwise."""
        return False  # pragma: no cover

    @property
    def is_running(self):
        r"""bool: True if the server is running."""
        if self.for_request:
            return True
        else:
            return self._is_running
    
    def wait_for_server(self, timeout=15.0):
        r"""Wait for a service to start running.

        Args:
            timeout (float, optional): Time (in seconds) that should be waited
                for the server to start. Defaults to 15.

        Raises:
            RuntimeError: If the time limit is reached and the server still
                hasn't started.

        """
        wait_on_function(lambda: self.is_running, timeout=timeout,
                         on_timeout="Server never started")

    def setup_server(self, *args, **kwargs):
        r"""Set up the machinery for receiving requests."""
        raise NotImplementedError  # pragma: no cover

    def setup_client(self, *args, **kwargs):
        r"""Set up the machinery for sending requests."""
        raise NotImplementedError  # pragma: no cover

    def set_log_level(self, log_level):
        r"""Set the logging level.

        Args:
            log_level (int): Logging level.

        """
        import logging
        logging.basicConfig(level=log_level)

    def start_server(self, remote_url=None, with_coverage=False,
                     log_level=None, model_repository=None):
        r"""Start the server.

        Args:
            remote_url (str optional): Address for the URL that remote
                integrations will use to connect to this server. Defaults to
                None and is set based on the YGGDRASIL_SERVICE_HOST_URL
                environment variable if it is set and is the local address
                otherwise.
            with_coverage (bool, optional): If True, the server is started
                with coverage. Defaults to False.
            log_level (int, optional): Level of log messages that should be
                printed. Defaults to None and is ignored.
            model_repository (str, optional): URL of directory in a Git
                repository containing YAMLs that should be added to the model
                registry. Defaults to None and is ignored.

        """
        if remote_url is None:
            remote_url = os.environ.get(_service_host_env, None)
        if remote_url is None:
            remote_url = self.address
        if model_repository is not None:
            repo_dir = self.registry.add_from_repository(model_repository)
            os.environ.setdefault(_service_repo_dir, repo_dir)
        os.environ.setdefault(_service_host_env, remote_url)
        if log_level is not None:
            self.set_log_level(log_level)
        if with_coverage:  # pragma: testing
            def handle_shutdown(sig, frame):
                sys.exit()
            signal.signal(_shutdown_signal, handle_shutdown)
            # try:
            #     from pytest_cov.embed import cleanup_on_signal
            #     cleanup_on_signal(_shutdown_signal)
            # except ImportError:  # pragma: debug
            #     pass
        self._is_running = True
        try:
            self.run_server()
        finally:
            self._is_running = False

    def run_server(self):
        r"""Begin listening for requests."""
        raise NotImplementedError  # pragma: no cover

    def respond(self, request, **kwargs):
        r"""Create a response to the request."""
        raise NotImplementedError  # pragma: no cover

    def shutdown(self, *args, **kwargs):
        r"""Shutdown the process from the server."""
        raise NotImplementedError  # pragma: no cover

    def process_request(self, request, **kwargs):
        r"""Process a request and return a response.

        Args:
            request (str): Serialized request that should be processed.
            **kwargs: Additional keyword arguments are passed to the respond
                method.

        Returns:
            str: Serialized response.

        """
        request = self.deserialize_request(request)
        response = self.respond(request, **kwargs)
        return self.serialize_response(response)

    def process_response(self, response):
        r"""Process a response.

        Args:
            response (str): Serialized response that should be processed.
        
        Returns:
            object: The deserialized, processed response.

        """
        return self.deserialize(response)

    def deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg (str): Message to deserialize.

        Returns:
            object: Deserialized message.

        """
        return json.loads(msg)

    def serialize(self, msg):
        r"""Serialize a message.

        Args:
            msg (object): Message to serialize.

        Returns:
            str: Serialized message.

        """
        return json.dumps(msg)

    def deserialize_request(self, request):
        r"""Deserialize a request message.

        Args:
            request (str): Serialized request.

        Returns:
            object: Deserialized request.

        """
        return self.deserialize(request)

    def serialize_response(self, response):
        r"""Serialize a response message.

        Args:
            request (object): Request to serialize.

        Returns:
            str: Serialized request.

        """
        return self.serialize(response)

    def call(self, *args, **kwargs):
        r"""Send a request."""
        raise NotImplementedError  # pragma: no cover

    def send_request(self, request, **kwargs):
        r"""Send a request.

        Args:
            request (object): Request to send.
            **kwargs: Additional keyword arguments are passed to the call
                method.

        Returns:
            object: Response.

        """
        request_str = self.serialize(request)
        assert(self.for_request)
        # if not self.for_request:
        #     x = self.__class__(self.name, *self._args,
        #                        **self._kwargs, for_request=True,
        #                        address=self.opp_address)
        # else:
        #     x = self
        return self.process_response(self.call(request_str, **kwargs))


class FlaskService(ServiceBase):
    r"""Flask based service."""

    service_type = 'flask'
    default_commtype = 'rest'
    default_address = 'http://localhost:{port}'
    default_port = int(os.environ.get("PORT", 5000))

    @classmethod
    def is_installed(cls):
        r"""bool: True if the class is fully installed, False otherwise."""
        try:
            import flask  # noqa: F401
            return True
        except ImportError:  # pragma: debug
            return False

    def __init__(self, *args, **kwargs):
        super(FlaskService, self).__init__(*args, **kwargs)
        # if str(self.port) not in self.address:
        #     parts = self.address.split(':')
        #     if parts[-1].strip('/').isdigit():
        #         self.port = int(parts[-1].strip('/'))
        if not self.address.endswith('/'):
            self.address += '/'

    def setup_server(self, *args, **kwargs):
        r"""Set up the machinery for receiving requests.

        Args:
            *args: Arguments are ignored.
            **kwargs: Keyword arguments are ignored.

        """
        from flask import Flask
        from flask import request
        from flask import jsonify
        self.queue = {}
        self.request = request
        self.jsonify = jsonify
        self.app = Flask(__name__)

        @self.app.route('/' + self.name, methods=['POST'])
        def _target(*req_args):
            return self.process_request(self.request.json, args=req_args)

    def setup_client(self, *args, **kwargs):
        r"""Set up the machinery for sending requests."""
        pass

    def set_log_level(self, log_level):
        r"""Set the logging level.

        Args:
            log_level (int): Logging level.

        """
        super(FlaskService, self).set_log_level(log_level)
        from flask.logging import default_handler
        werkzeug_logger = logging.getLogger('werkzeug')
        default_handler.setLevel(level=log_level)
        self.app.logger.setLevel(level=log_level)
        werkzeug_logger.setLevel(level=log_level)
        
    def run_server(self):
        r"""Begin listening for requests."""
        self.app.run(host='0.0.0.0', port=self.port)

    def shutdown(self):
        r"""Shutdown the process from the server."""
        # if not self.for_request:
        #     func = self.request.environ.get('werkzeug.server.shutdown')
        #     # Explicitly cleaning up the pytest coverage plugin is required
        #     # to ensure that the server methods are properly covered during
        #     # cleanup.
        #     try:
        #         from pytest_cov.embed import cleanup
        #         cleanup()
        #     except ImportError:  # pragma: debug
        #         pass
        #     if func is None:  # pragma: debug
        #         raise RuntimeError('Not running with the Werkzeug Server')
        #     func()  # pragma: no cover
        pass
        
    def deserialize(self, msg):
        r"""Deserialize a message.

        Args:
            msg (str): Message to deserialize.

        Returns:
            object: Deserialized message.

        """
        return msg  # should already be deserialized

    def serialize(self, msg):
        r"""Serialize a message.

        Args:
            msg (object): Message to serialize.

        Returns:
            str: Serialized message.

        """
        return msg  # should already be serialized
    
    def serialize_response(self, response):
        r"""Serialize a response message.

        Args:
            request (object): Request to serialize.

        Returns:
            str: Serialized request.

        """
        return self.jsonify(response)

    def call(self, request, **kwargs):
        r"""Send a request.

        Args:
            request (object): JSON serializable request.
            **kwargs: Keyword arguments are ignored.

        Returns:
            object: Response.

        """
        import requests
        try:
            r = requests.post(self.address + self.name, json=request)
            r.raise_for_status()
        except BaseException as e:
            raise ClientError(e)
        return r.json()
        

class RMQService(ServiceBase):
    r"""RabbitMQ based service."""

    service_type = 'rmq'
    default_commtype = 'rmq'
    default_port = 5672

    @classmethod
    def is_installed(cls):
        r"""bool: True if the class is fully installed, False otherwise."""
        from yggdrasil.communication.RMQComm import check_rmq_server
        return check_rmq_server()

    def _init_rmq(self, *args, **kwargs):
        from yggdrasil.communication.RMQComm import pika, get_rmq_parameters
        self.pika = pika
        if not self.address:
            kwargs['port'] = self.port
            self.address, _, _ = get_rmq_parameters(*args, **kwargs)
        self.queue = self.name
        # Unclear why using a non-default exchange prevents the server
        # from starting
        self.exchange = ''
        parameters = pika.URLParameters(self.address)
        self.connection = pika.BlockingConnection(parameters)
        self.channel = self.connection.channel()

    def setup_server(self, *args, **kwargs):
        r"""Set up the machinery for receiving requests.

        Args:
            *args: Arguments are used to initialize the RabbitMQ connections
                via _init_rmq.
            **kwargs: Keyword arguments are used to initialize the RabbitMQ
                connections via _init_rmq.

        """
        self._init_rmq(*args, **kwargs)
        if self.exchange:  # pragma: debug
            # self.channel.exchange_declare(exchange=self.exchange,
            #                               auto_delete=True)
            raise NotImplementedError("There is a bug when using the "
                                      "non-default exchange.")
        self.channel.queue_declare(queue=self.queue)
        self.channel.basic_qos(prefetch_count=1)
        self.consumer_tag = self.channel.basic_consume(
            queue=self.queue,
            on_message_callback=self._on_request)
        cb = functools.partial(self.shutdown, in_callback=True)
        self.channel.add_on_cancel_callback(cb)
    
    def setup_client(self, *args, **kwargs):
        r"""Set up the machinery for sending requests.

        Args:
            *args: Arguments are used to initialize the RabbitMQ connections
                via _init_rmq.
            **kwargs: Keyword arguments are used to initialize the RabbitMQ
                connections via _init_rmq.

        """
        self._init_rmq(*args, **kwargs)
        result = self.channel.queue_declare(queue='', exclusive=True)
        self.callback_queue = result.method.queue
        self.consumer_tag = self.channel.basic_consume(
            queue=self.callback_queue,
            on_message_callback=self._on_response,
            auto_ack=True)

    def run_server(self):
        r"""Listen for requests."""
        try:
            self.channel.start_consuming()
        except self.pika.exceptions.ChannelWrongStateError:  # pragma: debug
            pass

    def shutdown(self, in_callback=False):
        r"""Shutdown the process from the server."""
        if not self.channel:  # pragma: debug
            return
        if self.for_request:
            queue = self.callback_queue
            in_callback = False
        else:
            queue = self.queue
        if not in_callback:
            self.channel.basic_cancel(consumer_tag=self.consumer_tag)
            if not self.for_request:
                return
        self.channel.queue_delete(queue=queue)
        self.channel.close()
        self.channel = None
        self.connection.close()
        self.connection = None

    def _on_request(self, ch, method, props, body):
        response = self.process_request(body)
        ch.basic_publish(exchange=self.exchange,
                         routing_key=props.reply_to,
                         properties=self.pika.BasicProperties(
                             correlation_id=props.correlation_id),
                         body=response)
        ch.basic_ack(delivery_tag=method.delivery_tag)

    def _on_response(self, ch, method, props, body):
        if self.corr_id == props.correlation_id:
            self.response.set(body)

    @property
    def is_running(self):
        r"""bool: True if the server is running."""
        return (super(RMQService, self).is_running and bool(self.channel))

    def call(self, request, timeout=10.0, **kwargs):
        r"""Send a request.

        Args:
            request (str): Serialized request.
            timeout (float, optional): Time (in seconds) that should be waited
                for a response to be returned. Defaults to 10.
            **kwargs: Keyword arguments are ignored.

        Returns:
            str: Serialized response.

        """
        self.response = ValueEvent()
        self.corr_id = str(uuid.uuid4())
        try:
            self.channel.basic_publish(exchange=self.exchange,
                                       routing_key=self.queue,
                                       properties=self.pika.BasicProperties(
                                           reply_to=self.callback_queue,
                                           correlation_id=self.corr_id),
                                       body=request)
        except (self.pika.exceptions.UnroutableError,
                self.pika.exceptions.StreamLostError) as e:  # pragma: debug
            raise ClientError(e)
        
        def process_events():
            self.connection.process_data_events()
            return self.response.is_set()
        
        def client_error():
            raise ClientError("No response received")

        if not self.response.is_set():
            wait_on_function(
                process_events, timeout=timeout, polling_interval=0.5,
                on_timeout=client_error)
        return self.response.get()


def create_service_manager_class(service_type=None):
    r"""Create an integration manager service with the specified base.

    Args:
        service_type (ServiceBase, str, optional): Base class that should be
            used. Defaults to ('services', 'default_type') configuration
            options, if set, and 'flask' if not.

    Returns:
        type: Subclass of ServiceBase to handle starting/stopping integrations
            running as services.

    """
    if service_type is None:
        service_type = _default_service_type
    if isinstance(service_type, str):
        cls_map = {'flask': FlaskService, 'rmq': RMQService}
        service_type = cls_map[service_type]

    class IntegrationServiceManager(service_type):
        r"""Manager to track running integrations.

        Args:
            name (str): Name that should be used to initialize an address for
                the service. Defaults to 'ygg_integrations'.
            commtype (str, optional): Communicator type that should be used
                for the connections to services. Defaults to ('services',
                'default_comm') configuration option, if set, and None if not.
            is_app (bool, optional): If True, the service manager will be run
                as an app and will not be expected to be shut down by clients.
                Defaults to False.
            **kwargs: Additional keyword arguments are passed to the __init__
                method for the service_type class.

        """

        def __init__(self, name=None, commtype=None, is_app=False, **kwargs):
            if name is None:
                name = 'ygg_integrations'
            self.integrations = {}
            self.stopped_integrations = {}
            self.registry = IntegrationServiceRegistry()
            if commtype is None:
                commtype = _default_commtype
            self.commtype = commtype
            self.is_app = is_app
            super(IntegrationServiceManager, self).__init__(name, **kwargs)
            if self.commtype is None:
                self.commtype = self.default_commtype

        @property
        def client_id(self):
            r"""str: The ID that should be associated with a client on the
            current machine. Defaults to the configuration entry
            ('services', 'client_id') if it is set and is generated otherwise.
            """
            global _client_id
            if _client_id is None:
                _client_id = str(uuid.uuid4())
            return _client_id

        def send_request(self, name=None, yamls=None, action='start', **kwargs):
            r"""Send a request.

            Args:
                name (str, tuple, optional): A hashable object that will be
                    used to reference the integration. If not provided,
                    the yamls keyword will be used.
                yamls (list, str, optional): One or more YAML files defining
                    a network of models to run as a service. Defaults to None.
                action (str, optional): Action that is being requested.
                    Defaults to 'start'.
                **kwargs: Additional keyword arguments are included in the
                    request.

            """
            if isinstance(yamls, str):
                yamls = [yamls]
            if name is None:
                name = yamls
            request = dict(kwargs, name=name, yamls=yamls, action=action)
            request.setdefault('client_id', self.client_id)
            wait_for_complete = ((action in ['start', 'stop', 'shutdown'])
                                 and (service_type != RMQService))
            out = super(IntegrationServiceManager, self).send_request(request)
            if wait_for_complete and (out['status'] != 'complete'):
                def is_complete():
                    out.update(
                        super(IntegrationServiceManager, self).send_request(
                            request))
                    return (out['status'] == 'complete')
                wait_on_function(
                    is_complete, timeout=30, polling_interval=0.5,
                    on_timeout=f"Request did not complete: {request}")
            return out

        def setup_server(self, *args, **kwargs):
            r"""Set up the machinery for receiving requests."""
            super(IntegrationServiceManager, self).setup_server(*args, **kwargs)
            if service_type == FlaskService:
                @self.app.route('/')
                def landing_page():
                    from flask import render_template
                    import yaml
                    kwargs = {
                        'address': self.address,
                        'available': {
                            k: yaml.dump(v).splitlines()
                            for k, v in self.registry.registry.items()},
                        'running': {
                            k: {k2: v2.printStatus(return_str=True).splitlines()
                                for k2, v2 in v.items()}
                            for k, v in self.integrations.items()}}
                    out = render_template(
                        'service_manager_index.html', **kwargs)
                    return out
                
                from yggdrasil.communication import RESTComm
                RESTComm.add_comm_server_to_app(self.app)
            
        def stop_server(self):
            r"""Stop the server from the client-side."""
            assert(self.for_request)
            try:
                response = self.send_request(action='shutdown')
            except ClientError:  # pragma: debug
                return
            if response.get('pid', None):
                kill(response['pid'], _shutdown_signal)
            self.shutdown()

        def start_integration(self, client_id, x, yamls, **kwargs):
            r"""Start an integration if it is not already running.

            Args:
                client_id (str): ID associated with the client requesting the
                    integration be started.
                x (str, tuple): Hashable object that should be used to
                    identify the integration being started in the registry of
                    running integrations.
                yamls (list): Set of YAML specification files defining the
                    integration that should be run as as service.
                **kwargs: Additional keyword arguments are passed to get_runner.

            Returns:
                bool: True if the integration started, False otherwise.

            """
            integrations = self.integrations[client_id]
            stopped_integrations = self.stopped_integrations[client_id]
            if (x in integrations) and (not integrations[x].is_alive):
                if not self.stop_integration(client_id, x):
                    return False
            if x in stopped_integrations:
                if not self.stop_integration(client_id, x):
                    return False
                stopped_integrations.pop(x)
            if x not in integrations:
                partial_commtype = {'commtype': self.commtype}
                if self.commtype == 'rest':
                    partial_commtype['client_id'] = client_id
                integrations[x] = runner.get_runner(
                    yamls, complete_partial=True, as_service=True,
                    partial_commtype=partial_commtype, **kwargs)
                integrations[x].run(signal_handler=False)
            return True

        def _stop_integration(self, client_id, x):
            r"""Finish stopping an integration in a thread."""
            m = self.integrations[client_id].pop(x)
            if m.is_alive:
                m.terminate()
            m.atexit()

        def stop_integration(self, client_id, x):
            r"""Stop a running integration.

            Args:
                client_id (str): ID associated with the client requesting the
                    integration be stopped.
                x (str, tuple): Hashable object associated with the
                    integration service that should be stopped. If None,
                    all of the running integrations are stopped.

            Returns:
                bool: True if the integration has stopped.

            Raises:
                KeyError: If there is not a running integration associated
                    with the specified hashable object.

            """
            integrations = self.integrations[client_id]
            stopped_integrations = self.stopped_integrations[client_id]
            if x is None:
                return all(self.stop_integration(client_id, k)
                           for k in (list(integrations.keys())
                                     + list(stopped_integrations.keys())))
            if x in stopped_integrations:
                pass
            elif x not in integrations:
                raise KeyError(f"Integration defined by {x} not running")
            elif service_type == RMQService:
                self._stop_integration(client_id, x)
                return True
            else:
                mthread = threading.Thread(target=self._stop_integration,
                                           args=(client_id, x,), daemon=True)
                mthread.start()
                stopped_integrations[x] = mthread
            return not stopped_integrations[x].is_alive()

        def integration_info(self, client_id, x):
            r"""Get information about an integration and how to connect to it.

            Args:
                client_id (str): ID associated with the client requesting the
                    integration info.
                x (str, tuple): Hashable object associated with the
                    integration to get information on.

            Returns:
                dict: Information about the integration.

            Raises:
                KeyError: If there is not a running integration associated
                    with the specified hashable object.

            """
            integrations = self.integrations[client_id]
            if x not in integrations:  # pragma: debug
                raise KeyError(f"Integration defined by {x} not running")
            m = integrations[x].modeldrivers['dummy_model']
            out = m['instance'].service_partner
            name = 'dummy'
            if isinstance(x, str) and (not os.path.isfile(x)):
                name = x
            out.update(name=name,
                       args=name,
                       language='dummy')
            return out

        @property
        def is_running(self):
            r"""bool: True if the server is running."""
            if not super(IntegrationServiceManager, self).is_running:
                return False
            if self.for_request:
                try:
                    response = self.send_request(action='ping')
                    return response['status'] == 'running'
                except ClientError:
                    return False
            else:  # pragma: debug
                # This would only occur if a server calls is_running while it
                # is running (perhaps in a future callback?)
                return True

        def respond(self, request, **kwargs):
            r"""Create a response to the request.

            Args:
                request (dict): Request to respond to.
                **kwargs: Additional keyword arguments are ignored.

            Returns:
                dict: Response to the request.

            """
            name = None
            action = None
            yamls = None
            client_id = None
            try:
                name = request.pop('name')
                action = request.pop('action')
                yamls = request.pop('yamls')
                client_id = request.pop('client_id')
                if client_id is not None:
                    self.integrations.setdefault(client_id, {})
                    self.stopped_integrations.setdefault(client_id, {})
                if isinstance(name, list):
                    name = tuple(name)
                if action == 'start':
                    if not yamls:
                        reg = self.registry.registry.get(name, None)
                        if isinstance(name, tuple):
                            yamls = list(name)
                        elif isinstance(name, str) and os.path.isfile(name):
                            yamls = [name]
                        elif reg is not None:
                            yamls = reg['yamls']
                            for k, v in reg.items():
                                if k not in ['name', 'yamls']:
                                    request.setdefault(k, v)
                        else:  # pragma: debug
                            raise RuntimeError("No YAML files specified.")
                    if self.start_integration(client_id, name, yamls,
                                              **request):
                        response = dict(
                            self.integration_info(client_id, name),
                            status='complete')
                    else:
                        response = {'status': 'starting'}
                elif action == 'stop':
                    if self.stop_integration(client_id, name):
                        response = {'status': 'complete'}
                    else:
                        response = {'status': 'stopping'}
                elif action == 'shutdown':
                    if self.stop_integration(client_id, None):
                        if self.is_app:  # pragma: no cover
                            response = {'status': 'complete'}
                        else:
                            self.shutdown()
                            response = {'status': 'complete',
                                        'pid': os.getpid()}
                    else:
                        response = {'status': 'shutting down'}
                elif action == 'status':
                    response = {'status': 'done'}
                    if name is None:
                        fmt = ('Address: %s\n'
                               'Available Services:\n%s\n'
                               'Running Services:\n%s')
                        registry_str = '\t' + '\n\t'.join(
                            pprint.pformat(self.registry.registry).splitlines())
                        if client_id is None:
                            clients = list(self.integrations.keys())
                        else:
                            clients = [client_id]
                        running_str = ''
                        for cli in clients:
                            running_str += '\tClient %s\n' % cli
                            for k, v in self.integrations[cli].items():
                                running_str += '\t\t%s:\n\t\t\t%s' % (
                                    k, '\n\t\t\t'.join(v.printStatus(
                                        return_str=True).splitlines()))
                        args = (self.address, registry_str, running_str)
                        response['status'] = fmt % args
                    else:
                        assert(client_id is not None)
                        response['status'] = (
                            self.integrations[client_id][name].printStatus(
                                return_str=True))
                elif action == 'ping':
                    response = {'status': 'running'}
                else:
                    raise RuntimeError(f"Unsupported action: '{action}'")
            except BaseException as e:
                tb = traceback.format_exc()
                response = {'error': str(e), 'traceback': tb}
                if action == 'start':  # pragma: intermittent
                    self.respond({'name': name, 'action': 'stop', 'yamls': None,
                                  'client_id': client_id})
            return response

        def process_response(self, response):
            r"""Process a response.

            Args:
                response (str): Serialized response that should be processed.

            Returns:
                object: The deserialized, processed response.

            Raises:
                ServerError: If the response indicates there was an error on
                    the server during the creation of the response.

            """
            response = super(
                IntegrationServiceManager, self).process_response(response)
            if 'error' in response:
                raise ServerError('%s\n%s' % (response['traceback'],
                                              response['error']))
            return response

        def printStatus(self, level='info', return_str=False):
            r"""Print the status of the service manager including available
            and running services."""
            status = self.send_request(action='status')
            if return_str:
                msg, _ = self.logger.process(status['status'], {})
                return msg
            getattr(self.logger, level)(status['status'])
            
    return IntegrationServiceManager


def IntegrationServiceManager(service_type=None, **kwargs):
    r"""Start a management service to track running integrations.

    Args:
        service_type (ServiceBase, str, optional): Base class that should be
            used. Defaults to ('services', 'default_type') configuration
            options, if set, and 'flask' if not. If there is an address
            provided, the service type will be determined by parsing the
            address.
        **kwargs: Additional keyword arguments are used to intialized the
            manager class instance.

    """
    if service_type is None:
        if kwargs.get('address', None):
            if kwargs['address'].startswith('amqp://'):
                service_type = 'rmq'
            else:
                service_type = 'flask'
    cls = create_service_manager_class(service_type=service_type)
    return cls(**kwargs)


class IntegrationServiceRegistry(object):
    r"""Class for managing integration services.

    Args:
        filename (str, optional): File where the registry will be/is stored.
            Defaults to '~/.yggdrasil_services.yml'.

    """

    def __init__(self, filename=os.path.join('~', '.yggdrasil_services.yml')):
        self.filename = os.path.expanduser(filename)

    @property
    def registry(self):
        r"""dict: Existing registry of integrations."""
        return self.load()

    def load(self):
        r"""Load the dictionary of existing integrations that have been
        registered.

        Returns:
            dict: Existing registry of integrations.

        """
        out = {}
        if os.path.isfile(self.filename):
            with open(self.filename, 'r') as fd:
                out = yaml.safe_load(fd)
        return out

    def save(self, registry):
        r"""Save the registry to self.filename.

        Args:
            registry (dict): Dictionary of integrations to save.

        """
        with open(self.filename, 'w') as fd:
            yaml.dump(registry, fd)

    def load_collection(self, name):
        r"""Read a collection of integration registry entries from an YAML.

        Args:
            name (str): Full path to a YAML file containing one or more
                registry entries mapping between integration name and YAML
                specification files.

        Returns:
            dict: Loaded registry entries.

        """
        with open(name, 'r') as fd:
            out = yaml.safe_load(fd.read())
        assert(isinstance(out, dict))
        base_dir = os.path.dirname(name)
        for k in out.keys():
            v = out.get(k, [])
            out[k] = []
            for x in v:
                if not os.path.isabs(x):
                    x = os.path.join(base_dir, x)
                out[k].append(x)
        return out

    def remove(self, name):
        r"""Remove an integration service from the registry.

        Args:
            name (str): Name associated with the integration service that
                should be removed from the registry.

        Raises:
            KeyError: If there is not an integration service associated with
                the specified name.

        """
        registry = self.load()
        if os.path.isfile(name):
            names = list(self.load_collection(name).keys())
        else:
            names = [name]
        for k in names:
            if k not in registry:
                keys = list(self.registry.keys())
                raise KeyError(f"There is not an integration service "
                               f"registered under the name '{k}'. Existing "
                               f"services are {keys}")
            registry.pop(k)
        self.save(registry)

    def add_from_repository(self, model_repository, directory=None):
        r"""Add integration services to the registry from a repository of
        model YAMLs.

        Args:
            model_repository (str): URL of directory in a Git repository
                containing YAMLs that should be added to the model registry.
            directory (str, optional): Directory where services from the
                model_repository should be cloned. Defaults to
                '~/.yggdrasil_service'.

        Returns:
            str: The directory where the repositories were cloned.

        """
        from yggdrasil.yamlfile import clone_github_repo, prep_yaml
        if directory is None:
            directory = os.path.expanduser(
                os.path.join('~', '.yggdrasil_services'))
        yaml_dir = clone_github_repo(model_repository,
                                     local_directory=directory)
        yaml_files = (glob.glob(os.path.join(yaml_dir, '*.yaml'))
                      + glob.glob(os.path.join(yaml_dir, '*.yml')))
        for x in yaml_files:
            # Calling prep_yaml allows the model repositories to be cloned
            # in advance to circumvent th hold place on git cloning on the
            # service manager (these models are assumed to be vetted so
            # they do not pose a security risk).
            prep_yaml(x, directory_for_clones=directory)
            self.add(os.path.splitext(os.path.basename(x))[0], x)
        return directory

    def add(self, name, yamls=None, **kwargs):
        r"""Add an integration service to the registry.

        Args:
            name (str): Name that will be used to access the integration
                service when starting or stopping it.
            yamls (str, list): Set of one or more YAML specification files
                defining the integration.
            **kwargs: Additional keyword arguments are added to the new entry.

        Raises:
            ValueError: If there is already an integration with the specified
                name.

        """
        registry = self.load()
        if os.path.isfile(name):
            assert(not yamls)
            collection = {k: dict(kwargs, name=k, yamls=v)
                          for k, v in self.load_collection(name).items()}
        else:
            assert(yamls)
            collection = {name: dict(kwargs, name=name, yamls=yamls)}
        for k, v in collection.items():
            if (k in registry) and (registry[k] != v):
                old = pprint.pformat(registry[k])
                new = pprint.pformat(v)
                raise ValueError(f"There is an registry integration "
                                 f"associated with the name '{k}'. Remove "
                                 f"the registry entry before adding a new "
                                 f"one.\n"
                                 f"    Registry:\n{old}\n    New:\n{new}")
            registry[k] = v
        self.save(registry)


def validate_model_submission(fname):
    r"""Validate a YAML file according to the standards for submission to
    the yggdrasil model repository.

    Args:
        fname (str): YAML file to validate or directory in which to check
            each of the YAML files.

    """
    from yggdrasil import yamlfile, runner
    if isinstance(fname, list):
        for x in fname:
            validate_model_submission(x)
        return
    elif os.path.isdir(fname):
        files = sorted(glob.glob(os.path.join(fname, '*.yml'))
                       + glob.glob(os.path.join(fname, '*.yaml')))
        for x in files:
            validate_model_submission(x)
        return
    # 1-2. YAML syntax and schema
    yml = yamlfile.parse_yaml(fname, model_submission=True)
    # 3a. LICENSE
    repo_dir = yml['models'][0]['working_dir']
    patterns = ['LICENSE', 'LICENSE.*']
    for x in patterns:
        if ((glob.glob(os.path.join(repo_dir, x.upper()))
             or glob.glob(os.path.join(repo_dir, x.lower())))):
            break
    else:
        raise RuntimeError("Model repository does not contain a LICENSE file.")
    # 4. Run & validate
    runner.run(fname, validate=True)



import pytest
from tests.examples import TestExample as base_class


@pytest.mark.suite('mpi')
class TestExampleModelErrorWithIO(base_class):
    r"""Test the model_error_with_io example."""
    
    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "model_error_with_io"



import pytest
from tests.metaschema.datatypes.test_MetaschemaType import (
    TestMetaschemaType as base_class)
from tests.metaschema.properties.test_ArgsMetaschemaProperty import (
    ValidArgsClass4, InvalidArgsClass)


class TestInstanceMetaschemaType(base_class):
    r"""Test class for InstanceMetaschemaType class."""

    _mod = 'yggdrasil.metaschema.datatypes.InstanceMetaschemaType'
    _cls = 'InstanceMetaschemaType'

    @pytest.fixture(scope="class")
    def typedef_base(self):
        r"""dict: Base type definition."""
        return {'class': ValidArgsClass4,
                'args': ValidArgsClass4.valid_args,
                'kwargs': ValidArgsClass4.valid_kwargs}
    
    @pytest.fixture(scope="class")
    def value(self):
        r"""object: Test instance."""
        return ValidArgsClass4(*ValidArgsClass4.test_args,
                               **ValidArgsClass4.test_kwargs)

    @pytest.fixture(scope="class")
    def valid_encoded(self, python_class, typedef_base):
        r"""list: Encoded objects that are valid under this type."""
        return [dict(typedef_base,
                     type=python_class.name)]

    @pytest.fixture(scope="class")
    def valid_decoded(self, value):
        r"""list: Objects that are valid under this type."""
        return [value]

    @pytest.fixture(scope="class")
    def invalid_decoded(self):
        r"""list: Objects that are invalid under this type."""
        return [int(1), 'hello']

    @pytest.fixture(scope="class")
    def invalid_validate(self):
        r"""list: Objects that are invalid under this type."""
        return [None, InvalidArgsClass()]

    @pytest.fixture(scope="class")
    def compatible_objects(self, value):
        r"""list: Objects that are compatible with this type."""
        return [(value, value, None)]

    @pytest.fixture
    def nested_result(self, nested_approx):
        r"""Prepare value for comparison."""
        def nested_result_w(x):
            if isinstance(x, ValidArgsClass4):
                return x
            return nested_approx(x)
        return nested_result_w



import pytest
import os
import copy
import yaml
import importlib
from yggdrasil import constants
from yggdrasil.components import import_component
from yggdrasil.languages import get_language_ext
from yggdrasil.metaschema.datatypes import encode_type
from tests.examples import TestExample as base_class


_full_lang = sorted([k for k, v in constants.LANGUAGE_PROPERTIES.items()
                     if v['full_language']])
_typed_lang = sorted([k for k, v in constants.LANGUAGE_PROPERTIES.items()
                      if (v['full_language'] and v['is_typed'])])


@pytest.mark.suite("types", disabled=True, ignore="examples")
class TestExampleTypes(base_class):
    r"""Test the Types example."""

    parametrize_language = _full_lang
    # iter_flaky = [('c', 'instance', '*'), ('cpp', 'instance', '*')]

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "types"

    @pytest.fixture(scope="class")
    def example_module(self, example_name):
        r"""Python module associated with the test."""
        try:
            return importlib.import_module(f'tests.examples.{example_name}')
        except ImportError:
            return None

    @pytest.fixture(scope="class", autouse=True)
    def language(self, request, check_required_languages):
        r"""str: Language of the currect test."""
        check_required_languages([request.param])
        return request.param

    @pytest.fixture(scope="class")
    def yaml(self, example_name, language):
        r"""str: The full path to the yaml file for this example."""
        return os.path.join(os.path.dirname(__file__), 'types', 'types.yml')

    @pytest.fixture(scope="class", autouse=True)
    def typename(self, request):
        r"""str: Type being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def is_typed(self, language):
        r"""bool: True if the tested is for a typed language."""
        return (language in _typed_lang)

    @pytest.fixture(scope="class",
                    params=[{},
                            {'using_pointers': True},
                            {'using_generics': True},
                            {'using_pointers': True,
                             'dont_add_lengths': True},
                            {'using_pointers': True,
                             'length_prefix': True},
                            {'split_array': True}])
    def options(self, request):
        r"""Options used to create the test."""
        return request.param

    @pytest.fixture(scope="class")
    def using_pointers(self, options, is_typed):
        r"""bool: True if the test should be for pointers rather than explicit
        types."""
        out = options.get('using_pointers', False)
        if out and (not is_typed):
            pytest.skip("pointers only tested for typed languages")
        return out

    @pytest.fixture(scope="class")
    def using_generics(self, options, is_typed):
        r"""bool: True if the test should use generic types."""
        out = options.get('using_generics', False)
        if out and (not is_typed):
            pytest.skip("generic only tested for typed languages")
        return out

    @pytest.fixture(scope="class")
    def dont_add_lengths(self, options, language, is_typed, typename):
        r"""bool: True if the test should not add length variables."""
        out = options.get('dont_add_lengths', False)
        if out and not (
                is_typed and (language == 'c')
                and (typename in ['string', 'bytes', 'unicode'])):
            pytest.skip("dont_add_lengths only enabled for c string types")
        return out

    @pytest.fixture(scope="class")
    def length_prefix(self, options, language, is_typed, typename):
        r"""bool: True if length prefixes should be used instead of suffixes."""
        out = options.get('length_prefix', False)
        if out and not (
                is_typed and (language == 'c')
                and (typename in ['1darray', 'ndarray'])):
            pytest.skip("length_prefix only enabled for c arrays")
        return out

    @pytest.fixture(scope="class")
    def split_array(self, options, using_pointers, using_generics,
                    typename):
        r"""bool: True if arrays should be split across multiple variables."""
        out = options.get('split_array', False)
        if out and not (typename == 'array'):
            pytest.skip("split_array only enabled for array types")
        return out

    @classmethod
    def get_varstr(cls, vars_list, language, using_pointers=False,
                   length_prefix=False, dont_add_lengths=False):
        r"""Determine the vars string that should be used in the yaml.

        Args:
            vars_list (list): List of variables.
            language (str): Language being tested.
            using_pointers (bool, optional): If True and the tested
                language supports pointers, pointers will be used rather
                than explicit arrays. Defaults to False.
            length_prefix (bool, optional): If True, the length variables
                will be given prefixes instead of suffixes. Defaults to
                False.
            dont_add_lengths (bool, optional): If True, lengths will not
                be added to the definition or assignments. Defaults to
                False.

        Returns:
            str: Variable string.

        """
        out = []
        if length_prefix:
            length_fmt = 'length_%s'
            ndim_fmt = 'ndim_%s'
            shape_fmt = 'shape_%s'
        else:
            length_fmt = '%s_length'
            ndim_fmt = '%s_ndim'
            shape_fmt = '%s_shape'
        for v in vars_list:
            out.append(v['name'])
            typename = v['datatype']['type']
            if (language == 'c') and (not dont_add_lengths):
                if using_pointers:
                    if typename in ['string', 'bytes',
                                    'unicode', '1darray']:
                        out.append(length_fmt % v['name'])
                    elif typename in ['ndarray']:
                        out += [ndim_fmt % v['name'],
                                shape_fmt % v['name']]
                elif typename in ['string', 'bytes', 'unicode']:
                    out.append(length_fmt % v['name'])
        return ', '.join(out)
        
    @pytest.fixture(scope="class")
    def env(self, example_name, language, typename, using_pointers,
            using_generics, split_array, dont_add_lengths,
            length_prefix, example_module):
        r"""dict: Environment variables set for the test."""
        kwargs = {}
        assign_kws = {}
        if language in ['c', 'c++', 'cpp']:
            # dont_add_lengths is only valid for C/C++
            kwargs['dont_add_lengths'] = dont_add_lengths
            kwargs['use_length_prefix'] = length_prefix
            assign_kws.setdefault('dont_add_lengths', dont_add_lengths)
            assign_kws.setdefault('use_length_prefix', length_prefix)
        yaml_fields = {'vars': False, 'dtype': False}
        language_ext = get_language_ext(language)
        modelfile = os.path.join(os.path.dirname(__file__), example_name,
                                 'src', 'model' + language_ext)
        drv = import_component('model', language)
        if using_generics and drv.is_typed:
            testtype = {'type': 'any'}
        else:
            testdata = example_module.get_test_data(typename)
            testtype = encode_type(testdata)
            using_generics = False
        if split_array and (typename == 'array'):
            inputs = [{'name': 'x%d' % i, 'datatype': x} for i, x in
                      enumerate(copy.deepcopy(testtype['items']))]
            outputs = [{'name': 'y%d' % i, 'datatype': x} for i, x in
                       enumerate(copy.deepcopy(testtype['items']))]
        else:
            inputs = [{'name': 'x',
                       'datatype': copy.deepcopy(testtype)}]
            outputs = [{'name': 'y',
                        'datatype': copy.deepcopy(testtype)}]
        # Write the model
        function_contents = []
        for i, o in zip(inputs, outputs):
            if using_pointers and drv.is_typed:
                for k in ['shape', 'length']:
                    i['datatype'].pop(k, None)
                    o['datatype'].pop(k, None)
            function_contents += drv.write_assign_to_output(
                o, i, outputs_in_inputs=drv.outputs_in_inputs,
                **assign_kws)
        lines = drv.write_function_def(
            'model', function_contents=function_contents,
            inputs=copy.deepcopy(inputs),
            outputs=copy.deepcopy(outputs),
            outputs_in_inputs=drv.outputs_in_inputs,
            opening_msg='IN MODEL', closing_msg='MODEL EXIT',
            print_inputs=True, print_outputs=True, **kwargs)
        with open(modelfile, 'w') as fd:
            print(modelfile)
            print('\n'.join(lines))
            fd.write('\n'.join(lines))
        env = {}
        env['TEST_LANGUAGE'] = language
        env['TEST_LANGUAGE_EXT'] = language_ext
        env['TEST_TYPENAME'] = typename
        if (language in ['c', 'fortran']) and (not using_generics):
            yaml_fields['vars'] = True
            if typename in ['array', 'object']:
                yaml_fields['dtype'] = True
        if any(list(yaml_fields.values())):
            lines = []
            for io, io_vars in zip(['input', 'output'],
                                   [inputs, outputs]):
                lines += [io + 's:',
                          '  name: %s' % io]
                if yaml_fields['vars']:
                    lines.append(
                        '  vars: %s' % self.get_varstr(
                            io_vars, language,
                            using_pointers=using_pointers,
                            length_prefix=length_prefix,
                            dont_add_lengths=dont_add_lengths))
                if yaml_fields['dtype']:
                    if len(io_vars) == 1:
                        dtype = io_vars[0]['datatype']
                    else:
                        dtype = {'type': 'array',
                                 'items': [x['datatype'] for
                                           x in io_vars]}
                    lines.append('  datatype:')
                    for x in yaml.dump(dtype).splitlines():
                        if "units: ''" in x:
                            continue
                        lines.append('    ' + x)
            env['TEST_MODEL_IO'] = '\n    '.join(lines) + '\n'
        else:
            env['TEST_MODEL_IO'] = ''
        try:
            yield env
        finally:
            if os.path.isfile(modelfile):
                os.remove(modelfile)

    @pytest.fixture
    def check_results(self):
        r"""This should be overridden with checks for the result."""
        def check_results_w():
            pass
        return check_results_w



import os
from yggdrasil import tools
from yggdrasil.runner import run
tyk_dir = 'tyk5'

try:
    # Change to the solution directory
    old_dir = os.getcwd()
    if not old_dir.endswith(tyk_dir):
        os.chdir(os.path.join('solutions', tyk_dir))
    if not os.path.isdir('output'):
        os.mkdir('output')

    # Part 1: add copies to isolated light model
    tools.display_source_diff('../../yamls/shoot_v1.yml',
                              'yamls/shoot_copies.yml',
                              number_lines=True)

    # Part 2: run it & check output
    run(['yamls/shoot_copies.yml'], production_run=True)
    tools.display_source('output/height.txt')

finally:
    os.chdir(old_dir)



import pytest
import os
from yggdrasil import constants
from tests.drivers.test_ConnectionDriver import (
    TestOutputDriver as base_class)


_filetypes = sorted(constants.COMPONENT_REGISTRY['file']['subtypes'].keys())


class TestFileOutputDriver(base_class):
    r"""Test class for FileOutputDriver."""

    parametrize_filetype = _filetypes
    
    test_send_recv = None
    test_send_recv_nolimit = None
    test_send_recv_closed = None
    
    @pytest.fixture(scope="class")
    def component_subtype(self):
        r"""Subtype of component being tested."""
        return 'file_output'

    @pytest.fixture(scope="class")
    def filetype(self, request):
        r"""str: Name of the file type being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def commtype(self, filetype):
        r"""str: Name of the communicator being tested."""
        return filetype

    @pytest.fixture
    def instance_args(self, name, filepath):
        r"""Arguments for a new instance of the tested class."""
        return (name, filepath)

    @pytest.fixture(autouse=True)
    def filepath(self, name, ocomm_python_class):
        r"""str: Path to the test file."""
        out = os.path.abspath(
            f'{name}_input{ocomm_python_class._default_extension}')
        try:
            yield out
        finally:
            if os.path.isfile(out):
                os.remove(out)

    @pytest.fixture
    def recv_comm(self):
        r"""CommBase: communicator for receiving messages from the driver."""
        pytest.skip("recv_comm disabled for output files")

    @pytest.fixture
    def before_instance_started(self, send_comm):
        r"""Actions performed after teh instance is created, but before it
        is started."""
        def before_instance_started_w(x):
            pass
        return before_instance_started_w
    
    @pytest.fixture
    def after_instance_started(self, send_comm, testing_options):
        r"""Action taken after the instance is started, but before tests
        begin."""
        def after_instance_started_w(x):
            for x in testing_options['send']:
                flag = send_comm.send(x)
                assert(flag)
            flag = send_comm.send_eof()
            assert(flag)
        return after_instance_started_w

    @pytest.fixture
    def run_before_stop(self, instance):
        r"""Commands to run while the instance is running."""
        def run_before_stop_w():
            instance.wait(1.0)
        return run_before_stop_w
    
    @pytest.fixture
    def contents_to_read(self, testing_options):
        r"""str: Contents that should be read to the file."""
        return testing_options['contents']

    @pytest.fixture
    def assert_after_stop(self, assert_after_terminate, filepath,
                          testing_options, contents_to_read):
        r"""Assertions to make after stopping the driver instance."""
        def assert_after_stop_w():
            assert_after_terminate()
            assert(os.path.isfile(filepath))
            if testing_options.get('exact_contents', True):
                with open(filepath, 'rb') as fd:
                    data = fd.read()
                assert(data == contents_to_read)
        return assert_after_stop_w
    
    # These are disabled to prevent writting extraneous data
    @pytest.fixture(scope="class")
    def run_before_terminate(self):
        r"""Commands to run while the instance is running, before terminate."""
        def run_before_terminate_w():
            pass
        return run_before_terminate_w


class TestAsciiTableArrayOutputDriver(TestFileOutputDriver):
    r"""Test class for FileOutputDriver reading to a table as an array."""

    @pytest.fixture(scope="class")
    def filetype(self):
        r"""str: Name of the file type being tested."""
        return 'table'

    @pytest.fixture(scope="class")
    def options(self):
        r"""Arguments that should be provided when getting testing options."""
        return {'array_columns': True}



import pytest
from yggdrasil.communication import new_comm
from yggdrasil.communication import IPCComm, CommBase


_ipc_installed = IPCComm.IPCComm.is_installed(language='python')


@pytest.mark.skipif(not _ipc_installed, reason="IPC library not installed")
def test_queue():
    r"""Test creation/removal of queue."""
    mq = IPCComm.get_queue()
    key = str(mq.key)
    assert(CommBase.is_registered('ipc', key))
    IPCComm.IPCComm.unregister_comm(key, dont_close=True)
    with pytest.raises(KeyError):
        IPCComm.remove_queue(mq)
    IPCComm.IPCComm.register_comm(key, mq)
    IPCComm.remove_queue(mq)
    assert(not CommBase.is_registered('ipc', key))


@pytest.mark.skipif(not _ipc_installed, reason="IPC library not installed")
def test_ipcs():
    r"""Test list of ipc objects."""
    IPCComm.ipcs()

    
@pytest.mark.skipif(not _ipc_installed, reason="IPC library not installed")
def test_ipc_queues():
    r"""Test list of ipc queues."""
    IPCComm.ipc_queues()


@pytest.mark.skipif(not _ipc_installed, reason="IPC library not installed")
def test_ipcrm():
    r"""Test removal of ipc objects."""
    IPCComm.ipcrm()


@pytest.mark.skipif(not _ipc_installed, reason="IPC library not installed")
def test_ipcrm_queues():
    r"""Test removal of ipc queues."""
    IPCComm.ipcrm_queues()
    assert(len(IPCComm.ipc_queues()) == 0)
    mq = IPCComm.get_queue()
    assert(len(IPCComm.ipc_queues()) == 1)
    IPCComm.ipcrm_queues(str(mq.key))
    assert(len(IPCComm.ipc_queues()) == 0)


@pytest.mark.skipif(_ipc_installed, reason="IPC library installed")
def test_queue_not_installed():  # pragma: windows
    r"""Test return of get_queue if IPC library is not installed."""
    assert(IPCComm.get_queue() is None)


@pytest.mark.skipif(_ipc_installed, reason="IPC library installed")
def test_ipcs_not_isntalled():  # pragma: windows
    r"""Test return of ipcs if IPC library is not installed."""
    assert(IPCComm.ipcs() == '')

    
@pytest.mark.skipif(_ipc_installed, reason="IPC library installed")
def test_ipcrm_not_isntalled():  # pragma: windows
    r"""Test ipcrm if IPC library is not installed."""
    IPCComm.ipcrm()

    
@pytest.mark.skipif(_ipc_installed, reason="IPC library installed")
def test_ipcrm_queues_not_isntalled():  # pragma: windows
    r"""Test ipcrm_queues if IPC library is not installed."""
    IPCComm.ipcrm_queues()

    
@pytest.mark.skipif(_ipc_installed, reason="IPC library installed")
def test_not_running():  # pragma: windows
    r"""Test raise of an error if a IPC library is not installed."""
    comm_kwargs = dict(commtype='ipc', direction='send', reverse_names=True)
    with pytest.raises(RuntimeError):
        new_comm('test', **comm_kwargs)



import pytest
import os
import six
from yggdrasil import yamlfile, runner, demos
from yggdrasil.components import import_component
from tests import TestBase as base_class


_demo_dir = os.path.dirname(os.path.abspath(demos.__file__))


class DemoMeta(type):

    def __new__(meta, name, bases, class_dict):
        
        @pytest.fixture(scope="class",
                        params=list(class_dict['runs'].keys()))
        def run_name(self, request):
            r"""str: Name of the run."""
            return request.param

        class_dict['run_name'] = run_name
        return type.__new__(meta, name, bases, class_dict)


@pytest.mark.suite("demos", disabled=True)
@six.add_metaclass(DemoMeta)
class DemoTstBase(base_class):
    r"""Base class for running demos."""

    runs = {}

    @pytest.fixture(scope="class", params=[])
    def demo_name(self, request):
        r"""str: Name of demo being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def demo_directory(self, demo_name):
        r"""str: Directory containing the demo."""
        return os.path.join(_demo_dir, demo_name)

    @pytest.fixture(scope="class", params=[])
    def run_name(self, request):
        r"""str: Name of the run."""
        return request.param

    @pytest.fixture(scope="class")
    def yamls(self, run_name, demo_directory):
        r"""tuple: YAMLs required for the run."""
        return [os.path.join(demo_directory, x) for x in self.runs[run_name]]

    def test_run(self, run_name, yamls, check_required_languages):
        r"""Run the integration."""
        languages = []
        for x in yamlfile.parse_yaml(list(yamls))['model'].values():
            drv = import_component('model', x['driver'],
                                   without_schema=True)
            if drv.language not in languages:
                languages.append(drv.language)
        check_required_languages(languages)
        # Run
        r = runner.get_runner(yamls, namespace=run_name,
                              production_run=True)
        r.run()
        assert(not r.error_flag)
        del r

    @pytest.fixture(scope="class", autouse=True)
    def create_output_directory(self, demo_directory):
        r"""Create the output directory if it dosn't exist."""
        out_dir = os.path.join(demo_directory, 'output')
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)



import pytest
import copy
import numpy as np
from yggdrasil import units
from tests.metaschema.datatypes.test_ScalarMetaschemaType import (
    TestScalarMetaschemaType as base_class)


class TestOneDArrayMetaschemaType(base_class):
    r"""Test class for ArrayMetaschemaType class."""
    
    _mod = 'yggdrasil.metaschema.datatypes.ArrayMetaschemaType'
    _cls = 'OneDArrayMetaschemaType'

    @pytest.fixture(scope="class")
    def explicit(self):
        r"""bool: If True the type is explicit."""
        return False

    @pytest.fixture(scope="class")
    def subtype(self):
        r"""str: Scalar base type."""
        return "float"
    
    @pytest.fixture(scope="class")
    def shape(self):
        r"""int,tuple: Shape of scalar/array."""
        return 10
    
    @pytest.fixture(scope="class")
    def valid_decoded(self, value, valid_units, dtype):
        r"""list: Objects that are valid under this type."""
        out = [value]
        for x in valid_units:
            out.append(units.add_units(copy.deepcopy(out[0]), x))
        out.append(np.array([], dtype))
        return out


class TestNDArrayMetaschemaType(base_class):
    r"""Test class for ArrayMetaschemaType class with 2D array."""

    _mod = 'yggdrasil.metaschema.datatypes.ArrayMetaschemaType'
    _cls = 'NDArrayMetaschemaType'

    @pytest.fixture(scope="class")
    def explicit(self):
        r"""bool: If True the type is explicit."""
        return False

    @pytest.fixture(scope="class")
    def subtype(self):
        r"""str: Scalar base type."""
        return "float"
    
    @pytest.fixture(scope="class")
    def shape(self):
        r"""int,tuple: Shape of scalar/array."""
        return (4, 5)



import pytest
from tests.metaschema.properties.test_ArgsMetaschemaProperty import (
    TestArgsMetaschemaProperty as base_class)


class TestKwargsMetaschemaProperty(base_class):
    r"""Test class for KwargsMetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.KwargsMetaschemaProperty'
    _cls = 'KwargsMetaschemaProperty'
    
    @pytest.fixture(scope="class")
    def valid(self, valid_instances):
        r"""Objects that are valid."""
        return [(x, x.__class__.valid_kwargs) for x in valid_instances]

    @pytest.fixture(scope="class")
    def invalid(self, valid_instances):
        r"""Objects that are invalid."""
        return [(x, x.__class__.invalid_kwargs) for x in valid_instances]

    @pytest.fixture(scope="class")
    def valid_type(self, valid_classes):
        r"""Valid type."""
        return valid_classes[0].valid_kwargs
    
    @pytest.fixture(scope="class")
    def invalid_type(self, valid_classes):
        r"""Invalid type."""
        return valid_classes[0].invalid_kwargs



import pytest
from yggdrasil.metaschema import (
    get_validator, get_metaschema, MetaschemaTypeError)
from yggdrasil.metaschema.properties.MetaschemaProperty import (
    create_property)
from tests import TestClassBase as base_class


def test_dynamic():
    r"""Test dynamic creation of property."""

    def encode(instance, typedef=None):
        return None

    def validate(validator, value, instance, schema):
        return

    def compare(prop1, prop2):
        if not prop2:
            yield 'Test error'
        return

    new_prop = create_property('invalid', None, encode, validate, compare,
                               dont_register=True)
    assert(new_prop.encode('hello') is None)
    assert(list(new_prop.validate(None, None, None, None)) == [])
    assert(list(new_prop.compare(True, True)) == [])
    assert(list(new_prop.compare(True, False)) == ['Test error'])


class TestMetaschemaProperty(base_class):
    r"""Test class for MetaschemaProperty class."""
    
    _mod = 'yggdrasil.metaschema.properties.MetaschemaProperty'
    _cls = 'MetaschemaProperty'

    @pytest.fixture(scope="class")
    def valid(self):
        r"""Objects that are valid."""
        return []

    @pytest.fixture(scope="class")
    def invalid(self):
        r"""Objects that are invalid."""
        return []

    @pytest.fixture(scope="class")
    def encode_errors(self):
        r"""Object that enduce errors during encoding."""
        return []

    @pytest.fixture(scope="class")
    def valid_compare(self):
        r"""Objects that successfully compare."""
        return [(0, 0)]

    @pytest.fixture(scope="class")
    def invalid_compare(self):
        r"""Objects that do not successfully compare."""
        return [(0, 1)]

    @pytest.fixture(scope="class")
    def valid_normalize_schema(self):
        r"""Schemas for normalization."""
        return []

    @pytest.fixture(scope="class")
    def validator(self):
        r"""Schema validator."""
        return get_validator()(get_metaschema())
    
    def test_encode(self, python_class, valid):
        r"""Test encode method for the class."""
        for (instance, value) in valid:
            x = python_class.encode(instance)
            errors = list(python_class.compare(x, value))
            assert(not errors)
        if python_class.name == 'base':
            with pytest.raises(NotImplementedError):
                python_class.encode(None)

    def test_encode_errors(self, python_class, encode_errors):
        r"""Test errors raised by encode."""
        for instance in encode_errors:
            with pytest.raises(MetaschemaTypeError):
                python_class.encode(instance)

    def test_validate_valid(self, python_class, validator, valid):
        r"""Test validation method for the class on valid objects."""
        for (instance, value) in valid:
            schema = {python_class.name: value}
            errors = list(
                python_class.wrapped_validate(validator, value, instance, schema))
            assert(not errors)
        # Instances not of the associate type validate as true (skipped)
        if python_class.name == 'base':
            errors = list(
                python_class.wrapped_validate(validator, None, None, {}))
        assert(not errors)

    def test_validate_invalid(self, python_class, validator, invalid):
        r"""Test validation method for the class on invalid objects."""
        for (instance, value) in invalid:
            schema = {python_class.name: value}
            errors = list(
                python_class.wrapped_validate(validator, value, instance, schema))
            assert(errors)

    def test_compare_valid(self, python_class, valid_compare):
        r"""Test comparision method for the class on valid objects."""
        for x in valid_compare:
            errors = list(python_class.compare(*x))
            assert(not errors)

    def test_compare_invalid(self, python_class, invalid_compare):
        r"""Test comparision method for the class on invalid objects."""
        for x in invalid_compare:
            errors = list(python_class.compare(*x))
            assert(errors)

    def test_normalize_in_schema(self, python_class, valid_normalize_schema):
        r"""Test normalization in schema."""
        for x, y in valid_normalize_schema:
            assert(python_class.normalize_in_schema(x) == y)



def get_testing_options():
    r"""Get testing parameters for this example."""
    out = dict(
        output_dir='Output',
        output_files=['canopy_structure.txt',
                      'growth_rate.txt',
                      'light_intensity.txt',
                      'photosynthesis_rate.txt'],
        skip_check_results=True)
    return out



import os
from yggdrasil import import_as_function
from yggdrasil import tools, units
tyk_dir = 'tyk7'

try:
    # Change to the solution directory
    old_dir = os.getcwd()
    if not old_dir.endswith(tyk_dir):
        os.chdir(os.path.join('solutions', tyk_dir))
    if not os.path.isdir('output'):
        os.mkdir('output')

    # Part 1: import the weather model
    x = import_as_function('yamls/weather_alone.yml')
    x.model_info()
    print(x(123.0))
    x.stop()

    # Part 2: import the co2 model
    x = import_as_function('yamls/co2.yml')
    x.model_info()
    print(x(units.add_units(24.0, 'hrs'), units.add_units(2.9, 'cm')))
    x.stop()

finally:
    os.chdir(old_dir)



def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        temp_output_dir=True,
        input_files=['input.txt'],
        output_files=['output_hello.txt'])



import pytest
import os
from tests.examples import TestExample as base_class


@pytest.mark.extra_example
class TestExampleSaM(base_class):
    r"""Test the SaM example."""

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "SaM"

    @pytest.fixture
    def results(self, language):
        r"""list: Results that should be found in the output files."""
        # 1 + 2*n_languages
        if language == 'all':  # pragma: matlab
            s = 11
        elif language == 'all_nomatlab':  # pragma: no matlab
            s = 9
        else:
            s = 3
        return ['%d' % s]

    @pytest.fixture
    def output_files(self, tempdir):
        r"""list: Output files for the run."""
        return [os.path.join(tempdir, 'SaM_output.txt')]



import os
import numpy as np
from yggdrasil import units


def light(doy, height):
    r"""Compute the intensity of light.

    Args:
       doy (float): Day of year.
       height (float): Distance from ground in cm.

    Returns:
       float: Intensity of light in ergs cm^-2 s^-1.

    """
    # Define parameters that are static across a run
    amplitude = units.add_units(80.0, 'ergs cm^-3 s^-1')
    doy_offset = units.add_units(0.0, 'days')

    # Calculate intensity
    intensity = (
        amplitude * height *
        (1.0 + np.sin(2.0 * np.pi * (doy - doy_offset) /
                      units.add_units(365.0, 'days'))))
    
    # Check if model is running as a part of an yggdrasil integration
    with_yggdrasil = os.environ.get('YGG_SUBPROCESS', False)

    # If the model is running as part of an yggdrasil integration, import
    # the relevant yggdrasil routines and use the interface routine to
    # complete the connection defined in the YAML
    if with_yggdrasil:
        from yggdrasil.languages.Python.YggInterface import YggRpcClient
        weather_rpc = YggRpcClient('weather_light', global_scope=True)

        # Call the weather model
        flag, temp = weather_rpc.call(intensity)
        if not flag:
            raise Exception("Failed to call the weather model.")

        # Return both the intensity and temp
        return intensity, temp


    return intensity



import os
import numpy as np
import pandas as pd
from yggdrasil import units, tools
from yggdrasil.components import ComponentError, create_component


def get_test_data(transform=None):
    r"""Determine a test data set for the specified type.

    Returns:
        object: Example of specified datatype.

    """
    if transform is None:
        transform = os.environ['TEST_TRANSFORM']
    umol = b'\xce\xbcmol'.decode('utf-8')
    field_names = ['name', 'count', 'size']
    field_units = ['n/a', umol, 'cm']
    dtype = np.dtype(
        {'names': field_names,
         'formats': ['S5', 'i4', 'f8']})
    rows = [(b'one', np.int32(1), 1.0),
            (b'two', np.int32(2), 2.0),
            (b'three', np.int32(3), 3.0)]
    arr = np.array(rows, dtype=dtype)
    lst = [units.add_units(arr[n], u) for n, u
           in zip(field_names, field_units)]
    if transform == 'table':
        return list(rows[0])
    return lst


def check_received_data(transform, x_recv):
    r"""Check that the received message is equivalent to the
    test data for the specified type.

    Args:
        transform (str): Name of transform being tested.
        x_recv (object): Received object.

    Raises:
        AssertionError: If the received message is not equivalent
            to the received message.

    """
    try:
        t = create_component('transform', subtype=transform)
    except ComponentError:
        def t(x):
            return x
    x_sent = t(get_test_data(transform))
    print('RECEIVED:')
    tools.pprint_encoded(x_recv)
    print('EXPECTED:')
    tools.pprint_encoded(x_sent)
    if isinstance(x_sent, np.ndarray):
        np.testing.assert_array_equal(x_recv, x_sent)
    elif isinstance(x_sent, pd.DataFrame):
        pd.testing.assert_frame_equal(x_recv, x_sent)
    else:
        assert(x_recv == x_sent)



import os
import copy
import pprint
import jsonschema
import numpy as np
import yggdrasil
from yggdrasil import constants, units
from yggdrasil.metaschema.encoder import encode_json, decode_json
from yggdrasil.metaschema.properties import get_registered_properties
from yggdrasil.metaschema.datatypes import get_registered_types


_metaschema_fbase = '.ygg_metaschema.json'
_metaschema_fname = os.path.abspath(os.path.join(
    os.path.dirname(yggdrasil.__file__), _metaschema_fbase))
_metaschema = None
_validator = None
_base_schema = {u'$schema': u'http://json-schema.org/draft-04/schema'}


if os.path.isfile(_metaschema_fname):
    with open(_metaschema_fname, 'r') as fd:
        _metaschema = decode_json(fd)
    schema_id = _metaschema.get('id', _metaschema.get('$id', None))
    assert(schema_id is not None)
    _metaschema.setdefault('$schema', schema_id)
    _base_schema['$schema'] = _metaschema.get('$schema', schema_id)

        
_base_validator = jsonschema.validators.validator_for(_base_schema)

        
class MetaschemaTypeError(TypeError):
    r"""Error that should be raised when a class encounters a type it cannot handle."""
    pass


def create_metaschema(overwrite=False):
    r"""Create the meta schema for validating ygg schema.

    Args:
        overwrite (bool, optional): If True, the existing meta schema will be
            overwritten. If False and the metaschema exists, an error will be
            raised. Defaults to False.

    Returns:
        dict: Meta schema specifying rules for ygg type schema. This includes
            all original JSON schema rules with the addition of types and
            property definitions.

    Raises:
        RuntimeError: If the file already exists and overwrite is False.

    """
    if (not overwrite) and os.path.isfile(_metaschema_fname):
        raise RuntimeError("Metaschema file already exists.")
    out = copy.deepcopy(_base_validator.META_SCHEMA)
    out['title'] = "Ygg meta-schema for data type schemas"
    # TODO: Replace schema with a link to the metaschema in the documentation
    # del out['$schema']
    # Add properties
    for k, v in get_registered_properties().items():
        if v.schema is not None:
            assert(k not in out['properties'])
            out['properties'][k] = v.schema
    # Add types
    for k, v in sorted(get_registered_types().items()):
        if k not in out['definitions']['simpleTypes']['enum']:
            out['definitions']['simpleTypes']['enum'].append(k)
        for p in v.properties:
            assert(p in out['properties'])
    # Print
    print('Created metaschema')
    pprint.pprint(out)
    # Save it to a file
    with open(_metaschema_fname, 'w') as fd:
        encode_json(out, fd)
    return out


def get_metaschema():
    r"""Return the meta schema for validating ygg schema.

    Returns:
        dict: Meta schema specifying rules for ygg type schema. This includes
            all original JSON schema rules with the addition of types and
            property definitions.

    .. note:: This function should not be called at the module level as it can
              cause the metaschema (if it dosn't exist) to be generated before
              all of the necessary modules have been loaded.

    """
    global _metaschema
    if (_metaschema is None):
        _metaschema = create_metaschema()
    return copy.deepcopy(_metaschema)


def get_validator(overwrite=False, normalizers=None, **kwargs):
    r"""Return the validator that includes ygg expansion types.

    Args:
        overwrite (bool, optional): If True, the existing validator will be
            overwritten. Defaults to False.
        normalizers (dict, optional): Additional normalizers to add.
        **kwargs: Additional keyword arguments are passed to normalizer.create.

    Returns:
        jsonschema.IValidator: JSON schema validator.

    """
    from yggdrasil.metaschema import normalizer
    global _validator
    if (_validator is None) or overwrite:
        metaschema = get_metaschema()
        # Get set of validators
        all_validators = copy.deepcopy(_base_validator.VALIDATORS)
        for k, v in get_registered_properties().items():
            if (not v._replaces_existing):
                assert(k not in all_validators)
            all_validators[k] = v.wrapped_validate
        # Get set of datatypes
        type_checker = copy.deepcopy(_base_validator.TYPE_CHECKER)
        new_type_checkers = {}
        for k, v in get_registered_types().items():
            if (not v._replaces_existing):
                # Error raised on registration
                assert(k not in type_checker._type_checkers)
            new_type_checkers[k] = v.jsonschema_type_checker
        kwargs['type_checker'] = type_checker.redefine_many(
            new_type_checkers)
        # Get set of normalizers
        if normalizers is None:
            normalizers = {}
        # Use default base and update validators
        _validator = normalizer.create(meta_schema=metaschema,
                                       validators=all_validators,
                                       normalizers=normalizers, **kwargs)
        _validator._base_validator = _base_validator
    return _validator


def validate_schema(obj):
    r"""Validate a schema against the metaschema.

    Args:
        obj (dict): Schema to be validated.

    Raises:
        ValidationError: If the schema is not valid.

    """
    cls = get_validator()
    cls.check_schema(obj)


# def normalize_schema(obj):
#     r"""Normalize a schema against the metaschema.

#     Args:
#         obj (dict): Schema to be normalized.

#     Returns:
#         dict: Normalized schema.

#     """
#     cls = get_validator()
#     return cls.normalize_schema(obj)


def validate_instance(obj, schema, **kwargs):
    r"""Validate an instance against a schema.

    Args:
        obj (object): Object to be validated using the provided schema.
        schema (dict): Schema to use to validate the provided object.
        **kwargs: Additional keyword arguments are passed to validate.

    Raises:
        ValidationError: If the object is not valid.

    """
    cls = get_validator()
    cls.check_schema(schema)
    return cls(schema).validate(obj, **kwargs)


def normalize_instance(obj, schema, **kwargs):
    r"""Normalize an object using the provided schema.

    Args:
        obj (object): Object to be normalized using the provided schema.
        schema (dict): Schema to use to normalize the provided object.
        **kwargs: Additional keyword arguments are passed to normalize.
    
    Returns:
        object: Normalized instance.

    """
    cls = get_validator()
    cls.check_schema(schema)
    return cls(schema).normalize(obj, **kwargs)


def data2dtype(data):
    r"""Get numpy data type for an object.

    Args:
        data (object): Python object.

    Returns:
        np.dtype: Numpy data type.

    """
    data_nounits = units.get_data(data)
    if isinstance(data_nounits, np.ndarray):
        dtype = data_nounits.dtype
    elif isinstance(data_nounits, (list, dict, tuple)):
        raise MetaschemaTypeError
    elif isinstance(data_nounits,
                    np.dtype(constants.VALID_TYPES['bytes']).type):
        dtype = np.array(data_nounits).dtype
    else:
        dtype = np.array([data_nounits]).dtype
    return dtype


def definition2dtype(props):
    r"""Get numpy data type for a type definition.

    Args:
        props (dict): Type definition properties.
        
    Returns:
        np.dtype: Numpy data type.

    """
    typename = props.get('subtype', None)
    if typename is None:
        typename = props.get('type', None)
        if typename is None:
            raise KeyError('Could not find type in dictionary')
    if ('precision' not in props):
        if typename in constants.FLEXIBLE_TYPES:
            out = np.dtype((constants.VALID_TYPES[typename]))
        else:
            raise RuntimeError("Precision required for type: '%s'" % typename)
    elif typename == 'unicode':
        out = np.dtype((constants.VALID_TYPES[typename],
                        int(props['precision'] // 32)))
    elif typename in constants.FLEXIBLE_TYPES:
        out = np.dtype((constants.VALID_TYPES[typename],
                        int(props['precision'] // 8)))
    else:
        out = np.dtype('%s%d' % (constants.VALID_TYPES[typename],
                                 int(props['precision'])))
    return out


def type2numpy(typedef):
    r"""Convert a type definition into a numpy dtype.

    Args:
        typedef (dict): Type definition.

    Returns:
        np.dtype: Numpy data type.

    """
    out = None
    if ((isinstance(typedef, dict) and ('type' in typedef)
         and (typedef['type'] == 'array') and ('items' in typedef))):
        if isinstance(typedef['items'], dict):
            as_array = (typedef['items']['type'] in ['1darray', 'ndarray'])
            if as_array:
                out = definition2dtype(typedef['items'])
        elif isinstance(typedef['items'], (list, tuple)):
            as_array = True
            dtype_list = []
            field_names = []
            for i, x in enumerate(typedef['items']):
                if x['type'] not in ['1darray', 'ndarray']:
                    as_array = False
                    break
                dtype_list.append(definition2dtype(x))
                field_names.append(x.get('title', 'f%d' % i))
            if as_array:
                out = np.dtype(dict(names=field_names, formats=dtype_list))
    return out
